# -*- coding: utf-8 -*-

"""Command line interface for PyBEL.

Why does this file exist, and why not put this in ``__main__``? You might be tempted to import things from ``__main__``
later, but that will cause problems--the code will get executed twice:

- When you run ``python3 -m pybel`` python will execute``__main__.py`` as a script. That means there won't be any
  ``pybel.__main__`` in ``sys.modules``.
- When you import __main__ it will get executed again (as a module) because
  there's no ``pybel.__main__`` in ``sys.modules``.

.. seealso:: http://click.pocoo.org/5/setuptools/#setuptools-integration
"""

import json
import logging
import os
import sys
import time
from typing import List, Optional

import click
from click_plugins import with_plugins
from pkg_resources import iter_entry_points
from tqdm.autonotebook import tqdm

from .canonicalize import to_bel_script
from .constants import get_cache_connection
from .examples import (
    braf_graph,
    egf_graph,
    homology_graph,
    sialic_acid_graph,
    statin_graph,
)
from .exceptions import BELParserWarning
from .io import (
    from_bel_script,
    from_pickle,
    load,
    to_bel_commons,
    to_edgelist,
    to_graphml,
    to_gsea,
    to_neo4j,
    to_nodelink_file,
    to_pickle,
    to_sif,
    to_triples_file,
)
from .io.bel_commons_client import _get_host, _get_password, _get_user
from .manager import Manager
from .manager.database_io import to_database
from .manager.models import Edge, Namespace, Node
from .struct import (
    get_unused_annotations,
    get_unused_list_annotation_values,
    get_unused_namespaces,
)
from .struct.graph import BELGraph, WarningTuple
from .utils import get_corresponding_pickle_path

logger = logging.getLogger(__name__)


def _page(it):
    click.echo_via_pager("\n".join(map(str, it)))


connection_option = click.option(
    "-c",
    "--connection",
    default=get_cache_connection(),
    show_default=True,
    help="Database connection string.",
)

host_option = click.option(
    "--host",
    default=_get_host(),
    show_default=True,
    help="URL of BEL Commons.",
)
user_option = click.option(
    "--user",
    default=_get_user,
    show_default=True,
    prompt=True,
    help="User for BEL Commons",
)
password_option = click.option(
    "--password",
    default=_get_password,
    show_default=True,
    prompt=True,
    hide_input=True,
    help="Password for BEL Commons",
)


def _from_pickle_callback(ctx, param, file):
    path = file.name

    if not path.endswith(".bel"):
        return from_pickle(file)

    cache_path = get_corresponding_pickle_path(path)

    if not os.path.exists(cache_path):
        click.echo(
            "The BEL script {path} has not yet been compiled. First, try running the following command:\n\n "
            "pybel compile {path}\n".format(path=path),
        )
        sys.exit(1)

    return from_pickle(cache_path)


graph_pickle_argument = click.argument(
    "graph",
    metavar="path",
    type=click.File("rb"),
    callback=_from_pickle_callback,
)

graph_argument = click.argument(
    "graph",
    metavar="path",
    callback=lambda _, __, path: load(path),
)

LOG_FMT = "%(asctime)s %(levelname)-8s %(message)s"
LOG_DATEFMT = "%Y-%m-%d %H:%M:%S"


def _debug_callback(_ctx, _param, value):
    if not value:
        logging.basicConfig(level=logging.WARNING, format=LOG_FMT, datefmt=LOG_DATEFMT)
    elif value == 1:
        logging.basicConfig(level=logging.INFO, format=LOG_FMT, datefmt=LOG_DATEFMT)
    else:
        logging.basicConfig(level=logging.DEBUG, format=LOG_FMT, datefmt=LOG_DATEFMT)


verbose_option = click.option(
    "-v",
    "--verbose",
    count=True,
    callback=_debug_callback,
    expose_value=False,
)


@with_plugins(iter_entry_points("pybel.cli_plugins"))
@click.group(help="PyBEL CLI on {}".format(sys.executable))
@click.version_option()
@connection_option
@click.pass_context
def main(ctx, connection):
    """Command line interface for PyBEL."""
    ctx.obj = Manager(connection=connection)
    ctx.obj.bind()  # add the engine to the metadata and query property to the session


@main.command()
@click.argument("text")
@click.option("--pprint", is_flag=True)
def parse(text: str, pprint: bool):
    """Parse a single BEL statement and pring JSON output."""
    from .parser.parse_bel import parse as _parse

    click.echo(json.dumps(_parse(text), indent=2 if pprint else None))


@main.command()
@click.argument("path")
@click.option("--allow-naked-names", is_flag=True, help="Enable lenient parsing for naked names")
@click.option(
    "--disallow-nested",
    is_flag=True,
    help="Disable lenient parsing for nested statements",
)
@click.option(
    "--disallow-unqualified-translocations",
    is_flag=True,
    help="Disallow unqualified translocations",
)
@click.option("--no-identifier-validation", is_flag=True, help="Turn off identifier validation")
@click.option("--no-citation-clearing", is_flag=True, help="Turn off citation clearing")
@click.option(
    "-r",
    "--required-annotations",
    multiple=True,
    help="Specify multiple required annotations",
)
@click.option("--upgrade-urls", is_flag=True)
@click.option("--skip-tqdm", is_flag=True)
@verbose_option
@click.pass_obj
def compile(
    manager,
    path,
    allow_naked_names,
    disallow_nested,
    disallow_unqualified_translocations,
    no_identifier_validation,
    no_citation_clearing,
    required_annotations,
    upgrade_urls,
    skip_tqdm,
):
    """Compile a BEL script to a graph."""
    logger.debug("using connection: %s", manager.engine.url)

    click.secho("Compilation", fg="red", bold=True)
    if skip_tqdm:
        click.echo("```")
    graph = from_bel_script(
        path,
        manager=manager,
        use_tqdm=(not skip_tqdm),
        disallow_nested=disallow_nested,
        allow_naked_names=allow_naked_names,
        disallow_unqualified_translocations=disallow_unqualified_translocations,
        citation_clearing=(not no_citation_clearing),
        required_annotations=required_annotations,
        no_identifier_validation=no_identifier_validation,
        allow_definition_failures=True,
        upgrade_urls=upgrade_urls,
    )
    if skip_tqdm:
        click.echo("```")
    to_pickle(graph, get_corresponding_pickle_path(path))

    click.echo("")
    _print_summary(graph, ticks=skip_tqdm)

    sys.exit(0 if 0 == graph.number_of_warnings() else 1)


@main.command()
@graph_pickle_argument
def summarize(graph: BELGraph):
    """Summarize a graph."""
    _print_summary(graph)


def _print_summary(graph: BELGraph, ticks: bool = False):
    if not ticks:
        click.secho("Summary", fg="red", bold=True)
        graph.summarize()

    unused_namespaces = get_unused_namespaces(graph)
    if unused_namespaces:
        click.secho(
            "\nUnused Namespaces ({}/{})".format(len(unused_namespaces), len(graph.defined_namespace_keywords)),
            fg="red",
            bold=True,
        )
        if ticks:
            click.echo("```")
        for namespace in sorted(unused_namespaces):
            click.echo(namespace)
        if ticks:
            click.echo("```")

    unused_annotations = get_unused_annotations(graph)
    if unused_annotations:
        click.secho(
            "\nUnused Annotations ({}/{})".format(len(unused_annotations), len(graph.defined_annotation_keywords)),
            fg="red",
            bold=True,
        )
        if ticks:
            click.echo("```")
        for annotation in sorted(unused_annotations):
            click.echo(annotation)
        if ticks:
            click.echo("```")

    unused_annotation_list_values = get_unused_list_annotation_values(graph)
    if unused_annotation_list_values:
        click.secho("\nUnused List Annotation Values", fg="red", bold=True)
        if ticks:
            click.echo("```")
        for annotation, values in sorted(unused_annotation_list_values.items()):
            click.echo("{} ({}/{})".format(annotation, len(values), len(graph.annotation_list[annotation])))
            for value in sorted(values):
                click.echo("  {}".format(value))
        if ticks:
            click.echo("```")


@main.command()
@graph_pickle_argument
def warnings(graph: BELGraph):
    """List warnings from a graph."""
    echo_warnings_via_pager(graph.warnings)


@main.command()
@graph_pickle_argument
@click.pass_obj
def insert(manager, graph: BELGraph):
    """Insert a graph to the database."""
    to_database(graph, manager=manager, use_tqdm=True)


@main.command()
@graph_argument
@host_option
@user_option
@password_option
def upload(graph: BELGraph, host: str, user: str, password: str):
    """Upload a graph to BEL Commons."""
    resp = to_bel_commons(graph, host=host, user=user, password=password)
    resp.raise_for_status()
    click.echo(json.dumps(resp.json()))


@main.command()
@graph_pickle_argument
@click.option("--tsv", type=click.File("w"), help="Path to output a TSV file.")
@click.option("--edgelist", type=click.File("w"), help="Path to output a edgelist file.")
@click.option("--sif", type=click.File("w"), help="Path to output an SIF file.")
@click.option(
    "--gsea",
    type=click.File("w"),
    help="Path to output a GRP file for gene set enrichment analysis.",
)
@click.option("--graphml", help="Path to output a GraphML file. Use .graphml for Cytoscape.")
@click.option("--nodelink", type=click.File("w"), help="Path to output a node-link JSON file.")
@click.option("--bel", type=click.File("w"), help="Output canonical BEL.")
def serialize(graph: BELGraph, tsv, edgelist, sif, gsea, graphml, nodelink, bel):
    """Serialize a graph to various formats."""
    if tsv:
        logger.info("Outputting TSV to %s", tsv)
        to_triples_file(graph, tsv)

    if edgelist:
        logger.info("Outputting edgelist to %s", edgelist)
        to_edgelist(graph, edgelist)

    if sif:
        logger.info("Outputting SIF to %s", sif)
        to_sif(graph, sif)

    if graphml:
        logger.info("Outputting GraphML to %s", graphml)
        to_graphml(graph, graphml)

    if gsea:
        logger.info("Outputting GRP to %s", gsea)
        to_gsea(graph, gsea)

    if nodelink:
        logger.info("Outputting Nodelink JSON to %s", nodelink)
        to_nodelink_file(graph, nodelink)

    if bel:
        logger.info("Outputting BEL script to %s", bel)
        to_bel_script(graph, bel)


@main.command()
@graph_pickle_argument
@click.option(
    "--connection",
    default="http://localhost:7474/db/data/",
    help="Connection string for neo4j upload.",
)
@click.password_option()
def neo(graph: BELGraph, connection: str, password: str):
    """Upload to neo4j."""
    import py2neo

    neo_graph = py2neo.Graph(connection, password=password)
    to_neo4j(graph, neo_graph)


@main.command()
@click.pass_obj
@click.argument("agents", nargs=-1)
@click.option("--local", is_flag=True, help="Upload to local database.")
@host_option
def machine(manager: Manager, agents: List[str], local: bool, host: str):
    """Get content from the INDRA machine and upload to BEL Commons."""
    from indra.sources import indra_db_rest

    from pybel import from_indra_statements

    statements = indra_db_rest.get_statements(agents=agents)
    click.echo("got {} statements from INDRA".format(len(statements)))

    graph = from_indra_statements(
        statements,
        name="INDRA Machine for {}".format(", ".join(sorted(agents))),
        version=time.strftime("%Y%m%d"),
    )
    click.echo("built BEL graph with {} nodes and {} edges".format(graph.number_of_nodes(), graph.number_of_edges()))

    if 0 == len(graph):
        click.echo("not uploading empty graph")
        sys.exit(-1)

    if local:
        to_database(graph, manager=manager)
    else:
        resp = to_bel_commons(graph, host=host)
        resp.raise_for_status()


@main.group()
def manage():
    """Manage the database."""


@manage.command()
@click.confirmation_option()
@click.pass_obj
def drop(manager: Manager):
    """Drop the database."""
    manager.drop_all()


@manage.command()
@click.option("-v", "--debug", is_flag=True)
@click.pass_obj
def examples(manager: Manager, debug: bool):
    """Load examples to the database."""
    level = logging.DEBUG if debug else logging.INFO
    logging.basicConfig(level=level)
    logging.getLogger("pybel").setLevel(level)

    for graph in (
        sialic_acid_graph,
        statin_graph,
        homology_graph,
        braf_graph,
        egf_graph,
    ):
        if manager.has_name_version(graph.name, graph.version):
            click.echo("already inserted {}".format(graph))
            continue
        click.echo("inserting {}".format(graph))
        manager.insert_graph(graph, use_tqdm=True)


@manage.group()
def namespaces():
    """Manage namespaces."""


@namespaces.command()  # noqa:F811
@click.argument("url")
@click.pass_obj
def insert(manager: Manager, url: str):
    """Add a namespace by URL."""
    manager.get_or_create_namespace(url)


def _ls(manager: Manager, model_cls, model_id: int):
    if model_id:
        n = manager.session.query(model_cls).get(model_id)
        _page(n.entries)

    else:
        for n in manager.session.query(model_cls).order_by(model_cls.uploaded.desc()):
            click.echo("\t".join(map(str, (n.id, n.keyword, n.version, n.url))))


@namespaces.command()
@click.option("-u", "--url", help="Specific resource URL to list")
@click.option("-i", "--namespace-id", type=int, help="Specific resource URL to list")
@click.pass_obj
def ls(manager: Manager, url: Optional[str], namespace_id: Optional[int]):
    """List cached namespaces."""
    if url:
        n = manager.get_or_create_namespace(url)
        _page(n.entries)
    elif namespace_id is not None:
        _ls(manager, Namespace, namespace_id)
    else:
        click.echo_via_pager(
            "\n".join("{}\t{}\t{}".format(n.id, n.name, n.url) for n in manager.session.query(Namespace)),
        )


@namespaces.command()  # noqa:F811
@click.argument("url")
@click.pass_obj
def drop(manager: Manager, url: str):
    """Drop a namespace by URL."""
    manager.drop_namespace_by_url(url)


@manage.group()
def networks():
    """Manage networks."""


@networks.command()  # noqa:F811
@click.pass_obj
def ls(manager: Manager):
    """List network names, versions, and optionally, descriptions."""
    for n in manager.list_networks():
        click.echo("{}\t{}\t{}".format(n.id, n.name, n.version))


@networks.command()  # noqa:F811
@click.option("-n", "--network-id", type=int, help="Identifier of network to drop")
@click.option(
    "-y",
    "--yes",
    is_flag=True,
    help="Drop all networks without confirmation if no identifier is given",
)
@click.pass_obj
def drop(manager: Manager, network_id: Optional[int], yes):
    """Drop a network by its identifier or drop all networks."""
    if network_id:
        manager.drop_network_by_id(network_id)

    elif yes or click.confirm("Drop all networks?"):
        manager.drop_networks()


@manage.group()
def edges():
    """Manage edges."""


@edges.command()  # noqa:F811
@click.option("--offset", type=int)
@click.option("--limit", type=int, default=10)
@click.pass_obj
def ls(manager: Manager, offset: Optional[int], limit: Optional[int]):
    """List edges."""
    q = manager.session.query(Edge)

    if offset:
        q = q.offset(offset)

    if limit > 0:
        q = q.limit(limit)

    for e in q:
        click.echo(e.bel)


@manage.group()
def nodes():
    """Manage nodes."""


@nodes.command()
@click.pass_obj
def prune(manager: Manager):
    """Prune nodes not belonging to any edges."""
    nodes_to_delete = [
        node for node in tqdm(manager.session.query(Node), total=manager.count_nodes()) if not node.networks
    ]
    manager.session.delete(nodes_to_delete)
    manager.session.commit()


@manage.command()  # noqa:F811
@click.pass_obj
def summarize(manager: Manager):
    """Summarize the contents of the database."""
    click.echo("Networks: {}".format(manager.count_networks()))
    click.echo("Edges: {}".format(manager.count_edges()))
    click.echo("Nodes: {}".format(manager.count_nodes()))
    click.echo("Namespaces: {}".format(manager.count_namespaces()))
    click.echo("Namespaces entries: {}".format(manager.count_namespace_entries()))
    click.echo("Annotations: {}".format(manager.count_annotations()))
    click.echo("Annotation entries: {}".format(manager.count_annotation_entries()))


def echo_warnings_via_pager(warnings: List[WarningTuple], sep: str = "\t") -> None:
    """Output the warnings from a BEL graph with Click and the system's pager."""
    # Exit if no warnings
    if not warnings:
        click.echo("Congratulations! No warnings.")
        sys.exit(0)

    max_line_width = max(len(str(exc.line_number)) for _, exc, _ in warnings)

    max_warning_width = max(len(exc.__class__.__name__) for _, exc, _ in warnings)

    s1 = "{:>" + str(max_line_width) + "}" + sep
    s2 = "{:>" + str(max_warning_width) + "}" + sep

    def _make_line(path: str, exc: BELParserWarning):
        s = click.style(path, fg="cyan") + sep
        s += click.style(s1.format(exc.line_number), fg="blue", bold=True)
        s += click.style(
            s2.format(exc.__class__.__name__),
            fg=("red" if exc.__class__.__name__.endswith("Error") else "yellow"),
        )
        s += click.style(exc.line, bold=True) + sep
        s += click.style(str(exc))
        return s

    click.echo_via_pager(
        "\n".join(_make_line(path, exc) for path, exc, _ in warnings),
    )


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-

"""This module contains base exceptions that are shared through the package.

A message for "General Parser Failure" is displayed when a problem was caused due to an unforeseen error. The line
number and original statement are printed for the user to debug.
"""

from .utils import ensure_quotes


class PyBELWarning(Exception):
    """The base class for warnings during compilation from which PyBEL can recover."""


class BELParserWarning(PyBELWarning):
    """The base PyBEL parser exception, which holds the line and position where a parsing problem occurred."""

    def __init__(self, line_number: int, line: str, position: int, *args):
        """Initialize the BEL parser warning.

        :param line_number: The line number on which this warning occurred
        :param line: The content of the line
        :param position: The position within the line where the warning occurred
        """
        super().__init__(line_number, line, position, *args)
        self.line_number = line_number
        self.line = line
        self.position = position

    def __str__(self):
        return "General Parser Failure on line {} at pos {}: {}".format(self.line_number, self.position, self.line)


class BELSyntaxError(BELParserWarning, SyntaxError):
    """For general syntax errors."""


class InconsistentDefinitionError(BELParserWarning):
    """Base PyBEL error for redefinition."""

    def __init__(self, line_number: int, line: str, position: int, definition: str):
        super(InconsistentDefinitionError, self).__init__(line_number, line, position, definition)
        self.definition = definition

    def __str__(self):
        return "Tried to redefine {} with: {}".format(self.definition, self.line)


class RedefinedNamespaceError(InconsistentDefinitionError):
    """Raised when a namespace is redefined."""


class RedefinedAnnotationError(InconsistentDefinitionError):
    """Raised when an annotation is redefined."""


# Naming Warnings


class NameWarning(BELParserWarning):
    """The base class for errors related to nomenclature."""

    def __init__(self, line_number: int, line: str, position: int, name: str, *args):
        """Build a warning wrapping a given name."""
        super().__init__(line_number, line, position, name, *args)
        self.name = name


class NakedNameWarning(NameWarning):
    """Raised when there is an identifier without a namespace. Enable lenient mode to suppress."""

    def __str__(self):
        return '"{}" should be qualified with a valid namespace'.format(self.name)


class MissingDefaultNameWarning(NameWarning):
    """Raised if reference to value not in default namespace."""

    def __str__(self):
        return '"{}" is not in the default namespace'.format(self.name)


class NamespaceIdentifierWarning(NameWarning):
    """The base class for warnings related to namespace:name identifiers."""

    def __init__(self, line_number: int, line: str, position: int, namespace: str, name: str):
        """Initialize the namespace identifier warning.

        :param line_number: The line number of the line that caused the exception
        :param line: The line that caused the exception
        :param position: The line's position of the exception
        :param namespace: The namespace of the identifier
        :param name: The name of the identifier
        """
        super(NamespaceIdentifierWarning, self).__init__(line_number, line, position, name, namespace)
        self.namespace = namespace


class UndefinedNamespaceWarning(NamespaceIdentifierWarning):
    """Raised if reference made to undefined namespace."""

    def __str__(self):
        return '"{}" is not a defined namespace'.format(self.namespace)


class MissingNamespaceNameWarning(NamespaceIdentifierWarning):
    """Raised if reference to value not in namespace."""

    def __str__(self):
        return '"{}" is not in the {} namespace'.format(self.name, self.namespace)


class MissingNamespaceRegexWarning(NamespaceIdentifierWarning):
    """Raised if reference not matching regex."""

    def __str__(self):
        return """"{}" doesn't match the regex for {} namespace""".format(self.name, self.namespace)


class AnnotationWarning(BELParserWarning):
    """Base exception for annotation warnings."""

    def __init__(self, line_number, line, position, annotation, *args):
        """Build an AnnotationWarning.

        :param int line_number: The line number on which the warning occurred
        :param str line: The line on which the warning occurred
        :param int position: The position in the line that caused the warning
        :param str annotation: The annotation name that caused the warning
        """
        super(AnnotationWarning, self).__init__(line_number, line, position, annotation, *args)
        self.annotation = annotation


class UndefinedAnnotationWarning(AnnotationWarning):
    """Raised when an undefined annotation is used."""

    def __str__(self):
        return """"{}" is not defined""".format(self.annotation)


class MissingAnnotationKeyWarning(AnnotationWarning):
    """Raised when trying to unset an annotation that is not set."""

    def __str__(self):
        return """"{}" is not set, so it can't be unset""".format(self.annotation)


class AnnotationIdentifierWarning(AnnotationWarning):
    """Base exception for annotation:value pairs."""

    def __init__(self, line_number, line, position, annotation, value):
        super(AnnotationIdentifierWarning, self).__init__(line_number, line, position, annotation, value)
        self.value = value


class IllegalAnnotationValueWarning(AnnotationIdentifierWarning):
    """Raised when an annotation has a value that does not belong to the original set of valid annotation values."""

    def __str__(self):
        return '"{}" is not defined in the {} annotation'.format(self.value, self.annotation)


class MissingAnnotationRegexWarning(AnnotationIdentifierWarning):
    """Raised if annotation doesn't match regex."""

    def __str__(self):
        return """"{}" doesn't match the regex for {} annotation""".format(self.value, self.annotation)


# Provenance Warnings


class VersionFormatWarning(BELParserWarning):
    """Raised if the version string doesn't adhere to semantic versioning or ``YYYYMMDD`` format."""

    def __init__(self, line_number, line, position, version_string):
        super(VersionFormatWarning, self).__init__(line_number, line, position, version_string)
        self.version_string = version_string

    def __str__(self):
        return (
            'Version string "{}" neither is a date like YYYYMMDD nor adheres to semantic versioning.'
            " See http://semver.org/".format(self.version_string)
        )


class MetadataException(BELParserWarning):
    """Base exception for issues with document metadata."""

    def __str__(self):
        return 'Invalid metadata - "{}"'.format(self.line)


class MalformedMetadataException(MetadataException):
    """Raised when an invalid metadata line is encountered."""


class InvalidMetadataException(BELParserWarning):
    """Raised when an incorrect document metadata key is used.

    .. hint:: Valid document metadata keys are:

        - ``Authors``
        - ``ContactInfo``
        - ``Copyright``
        - ``Description``
        - ``Disclaimer``
        - ``Licenses``
        - ``Name``
        - ``Version``

    .. seealso:: BEL specification on the `properties section <http://openbel.org/language/web/version_1.0/
                 bel_specification_version_1.0.html#_properties_section>`_
    """

    def __init__(self, line_number, line, position, key, value):
        super(InvalidMetadataException, self).__init__(line_number, line, position, key, value)
        self.key = key
        self.value = value

    def __str__(self):
        return "Invalid document metadata key: {}".format(self.key)


class MissingMetadataException(BELParserWarning):
    """Raised when a BEL Script is missing critical metadata."""

    def __init__(self, line_number, line, position, key):
        super(MissingMetadataException, self).__init__(line_number, line, position, key)
        self.key = key

    def __str__(self):
        return "Missing required document metadata: {}".format(self.key)

    @staticmethod
    def make(key: str):
        """Build an instance of this class with auto-filled dummy values.

        Unlike normal classes, polymorphism on __init__ can't be used for exceptions when pickling/unpickling.
        """
        return MissingMetadataException(0, "", 0, key)


class InvalidCitationLengthException(BELParserWarning):
    """Base exception raised when the format for a citation is wrong."""


class CitationTooShortException(InvalidCitationLengthException):
    """Raised when a citation does not have the minimum of {type, name, reference}."""

    def __str__(self):
        return "Citation is missing required fields: {}".format(self.line)


class CitationTooLongException(InvalidCitationLengthException):
    """Raised when a citation has more than the allowed entries, {type, name, reference, date, authors, comments}."""

    def __str__(self):
        return "Citation contains too many entries: {}".format(self.line)


class MissingCitationException(BELParserWarning):
    """Raised when trying to parse a BEL statement, but no citation is currently set.

    This might be due to a previous error in the formatting of a citation.

    Though it's not a best practice, some BEL curators set other annotations before the citation. If this is the case
    in your BEL document, and you're *absolutely* sure that all ``UNSET`` statements are correctly written, you can use
    ``citation_clearing=True`` as a keyword argument in any of the IO functions in :func:`pybel.from_lines`,
    :func:`pybel.from_url`, or :func:`pybel.from_path`.
    """

    def __str__(self):
        return "Missing citation; can't add: {}".format(self.line)


class MissingSupportWarning(BELParserWarning):
    """Raised when trying to parse a BEL statement, but no evidence is currently set.

    All BEL statements must be qualified with evidence.

    If your data is serialized from a database and provenance information is not readily
    accessible, consider referencing the publication for the database, or a url pointing to the data from either
    a programmatically or human-readable endpoint.
    """

    def __str__(self):
        return "Missing evidence; can't add: {}".format(self.line)


class MissingAnnotationWarning(BELParserWarning):
    """Raised when trying to parse a BEL statement and a required annotation is not present."""

    def __init__(self, line_number, line, position, required_annotations):
        super(MissingAnnotationWarning, self).__init__(line_number, line, position, required_annotations)
        self.required_annotations = required_annotations

    def __str__(self):
        return "Missing annotations: {}".format(", ".join(sorted(self.required_annotations)))


class InvalidCitationType(BELParserWarning):
    """Raised when a citation is set with an incorrect type.

    .. hint:: Valid citation types include:

        - ``Book``
        - ``PubMed``
        - ``Journal``
        - ``Online Resource``
        - ``URL``
        - ``DOI``
        - ``Other``

    .. seealso:: OpenBEL wiki on `citations <https://wiki.openbel.org/display/BELNA/Citation>`_
    """

    def __init__(self, line_number, line, position, citation_type):
        super(InvalidCitationType, self).__init__(line_number, line, position, citation_type)
        self.citation_type = citation_type

    def __str__(self):
        return '"{}" is not a valid citation type'.format(self.citation_type)


class InvalidPubMedIdentifierWarning(BELParserWarning):
    """Raised when a citation is set whose type is ``PubMed`` but whose database identifier is not a valid integer."""

    def __init__(self, line_number, line, position, reference):
        super(InvalidPubMedIdentifierWarning, self).__init__(line_number, line, position, reference)
        self.reference = reference

    def __str__(self):
        return '"{}" is not a valid PubMed identifier'.format(self.reference)


# BEL Syntax Warnings


class MalformedTranslocationWarning(BELParserWarning):
    """Raised when there is a translocation statement without location information."""

    def __init__(self, line_number, line, position, tokens):
        super(MalformedTranslocationWarning, self).__init__(line_number, line, position, tokens)
        self.tokens = tokens

    def __str__(self):
        return "Unqualified translocation: {} {}".format(self.line, self.tokens)


class PlaceholderAminoAcidWarning(BELParserWarning):
    """Raised when an invalid amino acid code is given.

    One example might be the usage of X, which is a colloquial signifier for a truncation in a given position. Text
    mining efforts for knowledge extraction make this mistake often. X might also signify a placeholder amino acid.
    """

    def __init__(self, line_number, line, position, code):
        super(PlaceholderAminoAcidWarning, self).__init__(line_number, line, position, code)
        self.code = code

    def __str__(self):
        return "Placeholder amino acid found: {}".format(self.code)


class NestedRelationWarning(BELParserWarning):
    """Raised when encountering a nested statement.

    See our the docs for an explanation of why we explicitly do not support nested statements.
    """

    def __str__(self):
        return "Nesting is not supported. Split this statement: {}".format(self.line)


# Semantic Warnings


class InvalidEntity(BELParserWarning):
    """Raised when using a non-entity name for a name."""

    def __init__(self, line_number, line, position, namespace, name):
        super().__init__(line_number, line, position, namespace, name)
        self.namespace = namespace
        self.name = name

    def __str__(self):
        return "{}:{} should not be coded as an entity".format(self.namespace, ensure_quotes(self.name))


class InvalidFunctionSemantic(BELParserWarning):
    """Raised when an invalid function is used for a given node.

    For example, an HGNC symbol for a protein-coding gene YFG cannot be referenced as an miRNA with ``m(HGNC:YFG)``
    """

    def __init__(self, line_number, line, position, func, namespace, name, allowed_functions):
        super().__init__(line_number, line, position, func, namespace, name, allowed_functions)
        self.func = func
        self.namespace = namespace
        self.name = name
        self.allowed_functions = allowed_functions

    def __str__(self):
        return "{} {}:{} should be encoded as one of: {}".format(
            self.func,
            self.namespace,
            ensure_quotes(self.name),
            ", ".join(self.allowed_functions),
        )



# -*- coding: utf-8 -*-

"""Conversion functions for BEL graphs with INDRA.

After assembling a model with `INDRA <https://github.com/sorgerlab/indra>`_, a list of
:class:`indra.statements.Statement` can be converted to a :class:`pybel.BELGraph` with
:class:`indra.assemblers.pybel.PybelAssembler`.

.. code-block:: python

    from indra.assemblers.pybel import PybelAssembler
    import pybel

    stmts = [
        # A list of INDRA statements
    ]

    pba = PybelAssembler(
        stmts,
        name='Graph Name',
        version='0.0.1',
        description='Graph Description'
    )
    graph = pba.make_model()

    # Write to BEL file
    pybel.to_bel_path(belgraph, 'simple_pybel.bel')

.. warning::

    These functions are hard to unit test because they rely on a whole set of java dependencies and will likely
    not be for a while.
"""

import json
from typing import Any, List, Mapping, Optional, TextIO, Union

from networkx.utils import open_file

try:
    from pickle5 import load
except ImportError:
    from pickle import load

__all__ = [
    "from_indra_statements",
    "from_indra_statements_json",
    "from_indra_statements_json_file",
    "from_indra_pickle",
    "to_indra_statements",
    "to_indra_statements_json",
    "to_indra_statements_json_file",
    "from_biopax",
]


def from_indra_statements(
    stmts,
    name: Optional[str] = None,
    version: Optional[str] = None,
    description: Optional[str] = None,
    authors: Optional[str] = None,
    contact: Optional[str] = None,
    license: Optional[str] = None,
    copyright: Optional[str] = None,
    disclaimer: Optional[str] = None,
):
    """Import a model from :mod:`indra`.

    :param List[indra.statements.Statement] stmts: A list of statements
    :param name: The graph's name
    :param version: The graph's version. Recommended to use `semantic versioning <http://semver.org/>`_ or
                        ``YYYYMMDD`` format.
    :param description: The description of the graph
    :param authors: The authors of this graph
    :param contact: The contact email for this graph
    :param license: The license for this graph
    :param copyright: The copyright for this graph
    :param disclaimer: The disclaimer for this graph
    :rtype: pybel.BELGraph
    """
    from indra.assemblers.pybel import PybelAssembler

    if authors is None:
        authors = "INDRA"

    pba = PybelAssembler(
        stmts=stmts,
        name=name,
        version=version,
        description=description,
        authors=authors,
        contact=contact,
        license=license,
        copyright=copyright,
        disclaimer=disclaimer,
    )

    graph = pba.make_model()
    return graph


def from_indra_statements_json(stmts_json: List[Mapping[str, Any]], **kwargs):
    """Get a BEL graph from INDRA statements JSON.

    :rtype: BELGraph

    Other kwargs are passed to :func:`from_indra_statements`.
    """
    from indra.statements import stmts_from_json

    statements = stmts_from_json(stmts_json)
    return from_indra_statements(statements, **kwargs)


@open_file(0, mode="r")
def from_indra_statements_json_file(file, **kwargs):
    """Get a BEL graph from INDRA statements JSON file.

    :rtype: BELGraph

    Other kwargs are passed to :func:`from_indra_statements`.
    """
    return from_indra_statements_json(json.load(file), **kwargs)


def from_indra_pickle(path: str, **kwargs):
    """Import a model from :mod:`indra`.

    :param path: Path to pickled list of :class:`indra.statements.Statement`
    :rtype: pybel.BELGraph

    Other kwargs are passed to :func:`from_indra_statements`.
    """
    with open(path, "rb") as f:
        statements = load(f)

    return from_indra_statements(stmts=statements, **kwargs)


def to_indra_statements(graph):
    """Export this graph as a list of INDRA statements using the :py:class:`indra.sources.pybel.PybelProcessor`.

    :param pybel.BELGraph graph: A BEL graph
    :rtype: list[indra.statements.Statement]
    """
    from indra.sources.bel import process_pybel_graph

    pbp = process_pybel_graph(graph)
    return pbp.statements


def to_indra_statements_json(graph) -> List[Mapping[str, Any]]:
    """Export this graph as INDRA JSON list.

    :param pybel.BELGraph graph: A BEL graph
    """
    return [statement.to_json() for statement in to_indra_statements(graph)]


@open_file(1, mode="w")
def to_indra_statements_json_file(graph, path: Union[str, TextIO], indent: Optional[int] = 2, **kwargs):
    """Export this graph as INDRA statement JSON.

    :param pybel.BELGraph graph: A BEL graph
    :param path: A writable file or file-like

    Other kwargs are passed to :func:`json.dump`.
    """
    json.dump(to_indra_statements_json(graph), path, indent=indent, **kwargs)


def from_biopax(path: str, encoding: Optional[str] = None, **kwargs):
    """Import a model encoded in Pathway Commons `BioPAX <http://www.biopax.org/>`_ via :mod:`indra`.

    :param path: Path to a BioPAX OWL file
    :param encoding: The encoding passed to :func:`indra.sources.biopax.process_owl`.
        See https://github.com/sorgerlab/indra/pull/1199.
    :rtype: pybel.BELGraph

    Other kwargs are passed to :func:`from_indra_statements`.

    .. warning:: Not compatible with all BioPAX! See INDRA documentation.
    """
    from indra.sources.biopax import process_owl

    model = process_owl(path, encoding=encoding)
    return from_indra_statements(stmts=model.statements, **kwargs)



# -*- coding: utf-8 -*-

"""Tests for node selection functions."""

import unittest

from pybel import BELGraph
from pybel.dsl import Protein
from pybel.struct.filters import get_nodes_by_namespace


class TestNodeSelection(unittest.TestCase):
    """Tests for node selection functions."""

    def test_get_node_by_namespace(self):
        """Test getting nodes with a given namespace."""
        g = BELGraph()
        a = Protein(namespace="N1", name="a")
        b = Protein(namespace="N1", name="b")
        c = Protein(namespace="N2", name="c")
        d = Protein(namespace="N3", name="d")
        g.add_node_from_data(a)
        g.add_node_from_data(b)
        g.add_node_from_data(c)
        g.add_node_from_data(d)

        nodes = set(get_nodes_by_namespace(g, "N1"))

        self.assertIn(a, nodes)
        self.assertIn(b, nodes)
        self.assertNotIn(c, nodes)
        self.assertNotIn(d, nodes)

        nodes = set(get_nodes_by_namespace(g, ("N1", "N2")))

        self.assertIn(a, nodes)
        self.assertIn(b, nodes)
        self.assertIn(c, nodes)
        self.assertNotIn(d, nodes)



# -*- coding: utf-8 -*-

"""Constants for Hetionet."""

from ...dsl import Abundance, BiologicalProcess, Pathology, Population, Protein, Rna
from ...struct import BELGraph

HETIONET_PUBMED = "28936969"

##################
# Hetionet types #
##################

ANATOMY = "Anatomy"
GENE = "Gene"
PATHWAY = "Pathway"
BIOPROCESS = "Biological Process"
COMPOUND = "Compound"
SIDE_EFFECT = "Side Effect"
DISEASE = "Disease"
PHARMACOLOGICAL_CLASS = "Pharmacologic Class"
SYMPTOM = "Symptom"

DSL_MAP = {
    ANATOMY: "uberon",
    BIOPROCESS: "go",
    COMPOUND: "drugbank",
    DISEASE: "doid",
    GENE: "ncbigene",
    PATHWAY: "reactome",
    PHARMACOLOGICAL_CLASS: "drugcentral",
    SIDE_EFFECT: "umls",
    SYMPTOM: "mesh",
}

TYPE_BLACKLIST = {"Molecular Function", "Cellular Component"}

QUALIFIED_MAPPING = {
    (ANATOMY, Population, "upregulates", GENE, Rna, BELGraph.add_positive_correlation),
    (
        ANATOMY,
        Population,
        "downregulates",
        GENE,
        Rna,
        BELGraph.add_negative_correlation,
    ),
    (ANATOMY, Population, "expresses", GENE, Rna, BELGraph.add_correlation),
    (COMPOUND, Abundance, "resembles", COMPOUND, Abundance, BELGraph.add_association),
    (COMPOUND, Abundance, "upregulates", GENE, Protein, BELGraph.add_increases),
    (COMPOUND, Abundance, "downregulates", GENE, Protein, BELGraph.add_decreases),
    (COMPOUND, Abundance, "treats", DISEASE, Pathology, BELGraph.add_decreases),
    (COMPOUND, Abundance, "palliates", DISEASE, Pathology, BELGraph.add_decreases),
    (COMPOUND, Abundance, "causes", SIDE_EFFECT, Pathology, BELGraph.add_increases),
    (
        GENE,
        Protein,
        "interacts",
        GENE,
        Protein,
        BELGraph.add_binds,
    ),  # FIXME look into this
    (GENE, Protein, "regulates", GENE, Protein, BELGraph.add_regulates),
    (GENE, Rna, "covaries", GENE, Rna, BELGraph.add_correlation),
    (DISEASE, Pathology, "localizes", ANATOMY, Population, BELGraph.add_association),
    (DISEASE, Pathology, "associates", GENE, Protein, BELGraph.add_association),
    (DISEASE, Pathology, "upregulates", GENE, Rna, BELGraph.add_positive_correlation),
    (DISEASE, Pathology, "downregulates", GENE, Rna, BELGraph.add_negative_correlation),
    (DISEASE, Pathology, "presents", SYMPTOM, Pathology, BELGraph.add_association),
    (DISEASE, Pathology, "resembles", DISEASE, Pathology, BELGraph.add_association),
}
UNQUALIFIED_MAPPING = {
    (GENE, Protein, "participates", PATHWAY, BiologicalProcess, BELGraph.add_part_of),
    (
        GENE,
        Protein,
        "participates",
        BIOPROCESS,
        BiologicalProcess,
        BELGraph.add_part_of,
    ),
}

####################
# Drug action tags #
####################

ACTIVATES_ACTIONS = {
    "agonist",
    "potentiator",
    "inducer",
    "positive modulator",
    "partial agonist",
    "positive allosteric modulator",
    "activator",
    "stimulator",
}
INHIBITS_ACTIONS = {
    "inhibitor",
    "antagonist",
    "blocker",
    "partial antagonist",
    "inhibitor, competitive",
    "negative modulator",
    "negative allosteric modulator",
    "allosteric antagonist",
    "suppressor",
    "inhibitory allosteric modulator",
    "conversion inhibitor",
}
REGULATES_ACTIONS = {
    "modulator",
    "allosteric modulator",
}
BINDS_ACTIONS = {
    "substrate",
    "binder",
    "other/unknown",
    "ligand",
    "cofactor",
    "product of",
    "opener",
    "desensitize the target",
    "other",
    "unknown",
    "antibody",
    "binding",
    "adduct",
    "multitarget",
    "releasing agent",
}
TBH_ACTIONS = {}

ACTIVATES_ACTION_PAIRS = {
    ("activator", "substrate"),
    ("agonist", "binder"),
    ("agonist", "partial agonist"),
    ("inducer", "substrate"),
    ("agonist", "positive allosteric modulator"),
    ("positive allosteric modulator", "potentiator"),
}
INHIBITS_ACTION_PAIR = {
    ("agonist", "positive modulator"),
    ("allosteric antagonist", "antagonist"),
    ("antagonist", "blocker"),
    ("antagonist", "inhibitor"),
    ("antagonist", "multitarget"),
    ("antagonist", "substrate"),
    ("blocker", "inhibitor"),
    ("blocker", "modulator"),
    ("inhibitor", "modulator"),
    ("inhibitor", "multitarget"),
    ("inhibitor", "negative modulator"),
    ("inhibitor", "other"),
    ("inhibitor", "substrate"),
    ("negative modulator", "releasing agent"),
}
CONFLICTING_ACTION_PAIR = {
    ("inducer", "inhibitor", "substrate"),
    ("inducer", "inhibitor"),
    ("agonist", "antagonist"),
    ("antagonist", "partial agonist"),
    ("adduct", "inhibitor"),
    ("agonist", "antagonist", "modulator"),
}
UNINTERPRETABLE_ACTION_PAIR = {
    ("binder", "opener"),
}



# -*- coding: utf-8 -*-

"""The query manager for the database."""

import datetime
from typing import Iterable, List, Optional, Union

from sqlalchemy import and_, or_
from sqlalchemy.orm import aliased

from .lookup_manager import LookupManager
from .models import Author, Citation, Edge, Evidence, Namespace, NamespaceEntry, Node
from ..constants import CITATION_TYPE_PUBMED
from ..struct import BELGraph
from ..utils import parse_datetime

__all__ = [
    "QueryManager",
    "graph_from_edges",
]


def graph_from_edges(edges: Iterable[Edge], **kwargs) -> BELGraph:
    """Build a BEL graph from edges."""
    graph = BELGraph(**kwargs)
    graph.raise_on_missing_annotations = False

    for edge in edges:
        edge.insert_into_graph(graph)

    graph.raise_on_missing_annotations = True

    return graph


class QueryManager(LookupManager):
    """An extension to the Manager to make queries over the database."""

    def count_nodes(self) -> int:
        """Count the number of nodes in the database."""
        return self._count_model(Node)

    def query_nodes(
        self,
        bel: Optional[str] = None,
        type: Optional[str] = None,
        namespace: Optional[str] = None,
        name: Optional[str] = None,
    ) -> List[Node]:
        """Query nodes in the database.

        :param bel: BEL term that describes the biological entity. e.g. ``p(HGNC:APP)``
        :param type: Type of the biological entity. e.g. Protein
        :param namespace: Namespace keyword that is used in BEL. e.g. HGNC
        :param name: Name of the biological entity. e.g. APP
        """
        q = self.session.query(Node)

        if bel:
            q = q.filter(Node.bel.ilike(f"%{bel}%"))

        if type:
            q = q.filter(Node.type == type)

        if namespace or name:
            q = q.join(NamespaceEntry)

            if namespace:
                q = q.join(Namespace).filter(Namespace.keyword.ilike(namespace))

            if name:
                q = q.filter(NamespaceEntry.name.ilike(f"%{name}%"))

        return q

    def count_edges(self) -> int:
        """Count the number of edges in the database."""
        return self._count_model(Edge)

    def get_edges_with_citation(self, citation: Citation) -> List[Edge]:
        """Get the edges with the given citation."""
        return self.session.query(Edge).join(Evidence).filter(Evidence.citation == citation)

    def get_edges_with_citations(self, citations: Iterable[Citation]) -> List[Edge]:
        """Get edges with one of the given citations."""
        return self.session.query(Edge).join(Evidence).filter(Evidence.citation.in_(citations)).all()

    def search_edges_with_evidence(self, evidence: str) -> List[Edge]:
        """Search edges with the given evidence.

        :param evidence: A string to search evidences. Can use wildcard percent symbol (%).
        """
        return self.session.query(Edge).join(Evidence).filter(Evidence.text.like(evidence)).all()

    def search_edges_with_bel(self, bel: str) -> List[Edge]:
        """Search edges with given BEL.

        :param bel: A BEL string to use as a search
        """
        return self.session.query(Edge).filter(Edge.bel.like(bel))

    def get_edges_with_annotation(self, annotation: str, value: str) -> List[Edge]:
        """Search edges with the given annotation/value pair."""
        query = self.session.query(Edge).join(NamespaceEntry, Edge.annotations).join(Namespace)
        query = query.filter(Namespace.keyword == annotation).filter(NamespaceEntry.name == value)
        return query.all()

    @staticmethod
    def _add_edge_function_filter(query, edge_node_id, node_type):
        """See usage in self.query_edges."""
        return query.join(Node, edge_node_id == Node.id).filter(Node.type == node_type)

    def query_edges(
        self,
        bel: Optional[str] = None,
        source_function: Optional[str] = None,
        source: Union[None, str, Node] = None,
        target_function: Optional[str] = None,
        target: Union[None, str, Node] = None,
        relation: Optional[str] = None,
    ):
        """Return a query over the edges in the database.

        Usually this means that you should call ``list()`` or ``.all()`` on this result.

        :param bel: BEL statement that represents the desired edge.
        :param source_function: Filter source nodes with the given BEL function
        :param source: BEL term of source node e.g. ``p(HGNC:APP)`` or :class:`Node` object.
        :param target_function: Filter target nodes with the given BEL function
        :param target: BEL term of target node e.g. ``p(HGNC:APP)`` or :class:`Node` object.
        :param relation: The relation that should be present between source and target node.
        """
        if bel:
            return self.search_edges_with_bel(bel)

        query = self.session.query(Edge)

        if relation:
            query = query.filter(Edge.relation.like(relation))

        if source_function:
            source_node_table = aliased(Node)
            query = query.join(source_node_table, Edge.source_id == source_node_table.id).filter(
                source_node_table.type == source_function
            )

        if target_function:
            target_node_table = aliased(Node)
            query = query.join(target_node_table, Edge.target_id == target_node_table.id).filter(
                target_node_table.type == target_function
            )

        if source:
            if isinstance(source, str):
                source = self.query_nodes(bel=source)
                if source.count() == 0:
                    return []
                source = source.first()  # FIXME what if this matches multiple?
                query = query.filter(Edge.source == source)
            elif isinstance(source, Node):
                query = query.filter(Edge.source == source)
            else:
                raise TypeError("Invalid type of {}: {}".format(source, source.__class__.__name__))

        if target:
            if isinstance(target, str):
                targets = self.query_nodes(bel=target).all()
                target = targets[0]  # FIXME what if this matches multiple?
                query = query.filter(Edge.target == target)
            elif isinstance(target, Node):
                query = query.filter(Edge.target == target)
            else:
                raise TypeError("Invalid type of {}: {}".format(target, target.__class__.__name__))

        return query

    def query_citations(
        self,
        db: Optional[str] = None,
        db_id: Optional[str] = None,
        name: Optional[str] = None,
        author: Union[None, str, List[str]] = None,
        date: Union[None, str, datetime.date] = None,
        evidence_text: Optional[str] = None,
    ) -> List[Citation]:
        """Query citations in the database.

        :param db: Type of the citation. e.g. PubMed
        :param db_id: The identifier used for the citation. e.g. PubMed_ID
        :param name: Title of the citation.
        :param author: The name or a list of names of authors participated in the citation.
        :param date: Publishing date of the citation.
        :param evidence_text:
        """
        query = self.session.query(Citation)

        if author is not None:
            query = query.join(Author, Citation.authors)
            if isinstance(author, str):
                query = query.filter(Author.name.like(author))
            elif isinstance(author, Iterable):
                query = query.filter(Author.has_name_in(set(author)))
            else:
                raise TypeError

        if db and not db_id:
            query = query.filter(Citation.db.like(db))
        elif db_id and db:
            query = query.filter(Citation.db_id == db_id)
        elif db_id and not db:
            raise ValueError("reference specified without type")

        if name:
            query = query.filter(Citation.name.like(name))

        if date:
            if isinstance(date, datetime.date):
                query = query.filter(Citation.date == date)
            elif isinstance(date, str):
                query = query.filter(Citation.date == parse_datetime(date))

        if evidence_text:
            query = query.join(Evidence).filter(Evidence.text.like(evidence_text))

        return query.all()

    def query_edges_by_pubmed_identifiers(self, pubmed_identifiers: List[str]) -> List[Edge]:
        """Get all edges annotated to the documents identified by the given PubMed identifiers."""
        fi = and_(Citation.db == CITATION_TYPE_PUBMED, Citation.db_id.in_(pubmed_identifiers))
        return self.session.query(Edge).join(Evidence).join(Citation).filter(fi).all()

    @staticmethod
    def _edge_both_nodes(nodes: List[Node]):
        """Get edges where both the source and target are in the list of nodes."""
        node_ids = [node.id for node in nodes]

        return and_(
            Edge.source_id.in_(node_ids),
            Edge.target_id.in_(node_ids),
        )

    def query_induction(self, nodes: List[Node]) -> List[Edge]:
        """Get all edges between any of the given nodes (minimum length of 2)."""
        if len(nodes) < 2:
            raise ValueError("not enough nodes given to induce over")

        return self.session.query(Edge).filter(self._edge_both_nodes(nodes)).all()

    @staticmethod
    def _edge_one_node(nodes: List[Node]):
        """Get edges where either the source or target are in the list of nodes.

        Note: doing this with the nodes directly is not yet supported by SQLAlchemy

        .. code-block:: python

            return or_(
                Edge.source.in_(nodes),
                Edge.target.in_(nodes),
            )
        """
        node_ids = [node.id for node in nodes]

        return or_(
            Edge.source_id.in_(node_ids),
            Edge.target_id.in_(node_ids),
        )

    def query_neighbors(self, nodes: List[Node]) -> List[Edge]:
        """Get all edges incident to any of the given nodes."""
        return self.session.query(Edge).filter(self._edge_one_node(nodes)).all()



# -*- coding: utf-8 -*-

"""Conversion functions for BEL graphs with bytes and Python pickles."""

import gzip
from io import BytesIO
from typing import BinaryIO, Union

from networkx.utils import open_file

from .utils import raise_for_not_bel, raise_for_old_graph
from ..struct.graph import BELGraph

try:
    import pickle5 as pickle
except ImportError:
    import pickle

__all__ = [
    "to_bytes",
    "from_bytes",
    "to_bytes_gz",
    "from_bytes_gz",
    "to_pickle",
    "to_pickle_gz",
    "from_pickle",
    "from_pickle_gz",
]


def to_bytes(graph: BELGraph, protocol: int = pickle.HIGHEST_PROTOCOL) -> bytes:
    """Convert a graph to bytes with pickle.

    Note that the pickle module has some incompatibilities between Python 2 and 3. To export a universally importable
    pickle, choose 0, 1, or 2.

    :param graph: A BEL graph
    :param protocol: Pickling protocol to use. Defaults to ``HIGHEST_PROTOCOL``.

    .. seealso:: https://docs.python.org/3.6/library/pickle.html#data-stream-format
    """
    raise_for_not_bel(graph)
    return pickle.dumps(graph, protocol=protocol)


def from_bytes(bytes_graph: bytes, check_version: bool = True) -> BELGraph:
    """Read a graph from bytes (the result of pickling the graph).

    :param bytes_graph: File or filename to write
    :param check_version: Checks if the graph was produced by this version of PyBEL
    """
    graph = pickle.loads(bytes_graph)

    raise_for_not_bel(graph)
    if check_version:
        raise_for_old_graph(graph)

    return graph


def to_bytes_gz(graph: BELGraph, protocol: int = pickle.HIGHEST_PROTOCOL) -> bytes:
    """Convert a graph to gzipped bytes with pickle.

    :param graph: A BEL graph
    :param protocol: Pickling protocol to use. Defaults to ``HIGHEST_PROTOCOL``.
    """
    io = BytesIO()
    with gzip.open(io, mode="wb") as file:
        pickle.dump(graph, file, protocol=protocol)
    return io.getvalue()


def from_bytes_gz(bytes_graph: bytes) -> BELGraph:
    """Read a graph from gzipped bytes (the result of pickling the graph).

    :param bytes_graph: File or filename to write
    """
    with gzip.GzipFile(fileobj=BytesIO(bytes_graph), mode="rb") as file:
        return pickle.load(file)


@open_file(1, mode="wb")
def to_pickle(graph: BELGraph, path: Union[str, BinaryIO], protocol: int = pickle.HIGHEST_PROTOCOL) -> None:
    """Write this graph to a pickle file.

    Note that the pickle module has some incompatibilities between Python 2 and 3. To export a universally importable
    pickle, choose 0, 1, or 2.

    :param graph: A BEL graph
    :param path: A path or file-like
    :param protocol: Pickling protocol to use. Defaults to ``HIGHEST_PROTOCOL``.

    .. seealso:: https://docs.python.org/3.6/library/pickle.html#data-stream-format
    """
    raise_for_not_bel(graph)
    pickle.dump(graph, path, protocol)


def to_pickle_gz(graph: BELGraph, path: str, protocol: int = pickle.HIGHEST_PROTOCOL) -> None:
    """Write this graph to a gzipped pickle file."""
    with gzip.open(path, "wb") as file:
        to_pickle(graph, file, protocol=protocol)


@open_file(0, mode="rb")
def from_pickle(path: Union[str, BinaryIO], check_version: bool = True) -> BELGraph:
    """Read a graph from a pickle file.

    :param path: File or filename to read. Filenames ending in .gz or .bz2 will be uncompressed.
    :param bool check_version: Checks if the graph was produced by this version of PyBEL
    """
    graph = pickle.load(path)

    raise_for_not_bel(graph)
    if check_version:
        raise_for_old_graph(graph)

    return graph


def from_pickle_gz(path: str) -> BELGraph:
    """Read a graph from a gzipped pickle file."""
    with gzip.open(path, "rb") as file:
        return from_pickle(file)



# -*- coding: utf-8 -*-

"""Mocks for PyBEL testing."""

import itertools as itt
import os
from unittest import mock

from .constants import bel_dir_path, belanno_dir_path, belns_dir_path
from .utils import get_uri_name

__all__ = [
    "MockResponse",
    "MockSession",
    "mock_bel_resources",
]

_responses = [
    ("go.belns", os.path.join(belns_dir_path, "go-names.belns")),
    (
        "hgnc-human-genes-20170725.belns",
        os.path.join(belns_dir_path, "hgnc-names.belns"),
    ),
    ("chebi-20170725.belns", os.path.join(belns_dir_path, "chebi-names.belns")),
    (
        "species-taxonomy-id-20170511.belanno",
        os.path.join(belanno_dir_path, "species-taxonomy-id.belanno"),
    ),
    (
        "confidence-1.0.0.belanno",
        os.path.join(belanno_dir_path, "confidence-1.0.0.belanno"),
    ),
]


class MockResponse:
    """See http://stackoverflow.com/questions/15753390/python-mock-requests-and-the-response."""

    def __init__(self, url_to_mock: str):
        """Build a mock for the requests Response object."""
        _r = [
            (".belns", os.path.join(belns_dir_path, get_uri_name(url_to_mock))),
            (".belanno", os.path.join(belanno_dir_path, get_uri_name(url_to_mock))),
            (".bel", os.path.join(bel_dir_path, get_uri_name(url_to_mock))),
        ]

        self.path = None
        for suffix, path in itt.chain(_responses, _r):
            if url_to_mock.endswith(suffix):
                self.path = path
                break

        if self.path is None:
            raise ValueError("missing file")

        if not os.path.exists(self.path):
            raise ValueError("file doesn't exist: {}".format(self.path))

    def iter_lines(self):
        """Iterate the lines of the mock file."""
        with open(self.path, "rb") as file:
            yield from file

    def raise_for_status(self):
        """Mock raising an error, by not doing anything at all."""


class MockSession:
    """Patches the session object so requests can be redirected through the filesystem without rewriting BEL files."""

    def mount(self, prefix, adapter):
        """Mock mounting an adapter by not doing anything."""

    @staticmethod
    def get(url: str):
        """Mock getting a URL by returning a mock response."""
        return MockResponse(url)

    def close(self):
        """Mock closing a connection by not doing anything."""


mock_bel_resources = mock.patch("bel_resources.utils.requests.Session", side_effect=MockSession)



# -*- coding: utf-8 -*-

"""This module contains functions that mutate or make transformations on a network."""

from . import (
    collapse,
    deletion,
    expansion,
    induction,
    induction_expansion,
    inference,
    metadata,
    utils,
)
from .collapse import *
from .deletion import *
from .expansion import *
from .induction import *
from .induction_expansion import *
from .inference import *
from .inference import transfer
from .inference.transfer import *
from .metadata import *
from .utils import *

__all__ = [k for k in locals() if not k.startswith("_")]



# -*- coding: utf-8 -*-

"""Summary functions for errors and warnings encountered during the compilation of BEL script."""

import typing
from collections import Counter, defaultdict
from typing import Iterable, List, Mapping, Set

from ..filters.edge_predicates import edge_has_annotation
from ..graph import BELGraph, WarningTuple
from ...constants import ANNOTATIONS
from ...exceptions import (
    BELSyntaxError,
    MissingNamespaceNameWarning,
    MissingNamespaceRegexWarning,
    NakedNameWarning,
)

__all__ = [
    "get_syntax_errors",
    "count_error_types",
    "count_naked_names",
    "get_naked_names",
    "calculate_incorrect_name_dict",
    "calculate_error_by_annotation",
]


def get_syntax_errors(graph: BELGraph) -> List[WarningTuple]:
    """List the syntax errors encountered during compilation of a BEL script."""
    return [(path, exc, an) for path, exc, an in graph.warnings if isinstance(exc, BELSyntaxError)]


def count_error_types(graph: BELGraph) -> typing.Counter[str]:
    """Count the occurrence of each type of error in a graph.

    :return: A Counter of {error type: frequency}
    """
    return Counter(exc.__class__.__name__ for _, exc, _ in graph.warnings)


def _naked_names_iter(graph: BELGraph) -> Iterable[str]:
    """Iterate over naked name warnings from a graph."""
    for _, exc, _ in graph.warnings:
        if isinstance(exc, NakedNameWarning):
            yield exc.name


def count_naked_names(graph: BELGraph) -> typing.Counter[str]:
    """Count the frequency of each naked name (names without namespaces).

    :return: A Counter from {name: frequency}
    """
    return Counter(_naked_names_iter(graph))


def get_naked_names(graph: BELGraph) -> Set[str]:
    """Get the set of naked names in the graph."""
    return set(_naked_names_iter(graph))


def _iterate_namespace_name(graph: BELGraph) -> Iterable[typing.Tuple[str, str]]:
    for _, exc, _ in graph.warnings:
        if not isinstance(exc, (MissingNamespaceNameWarning, MissingNamespaceRegexWarning)):
            continue
        yield exc.namespace, exc.name


def calculate_incorrect_name_dict(graph: BELGraph) -> Mapping[str, List[str]]:
    """Get missing names grouped by namespace."""
    missing = defaultdict(list)

    for namespace, name in _iterate_namespace_name(graph):
        missing[namespace].append(name)

    return dict(missing)


def calculate_error_by_annotation(graph: BELGraph, annotation: str) -> Mapping[str, List[str]]:
    """Group error names by a given annotation."""
    results = defaultdict(list)

    for _, exc, ctx in graph.warnings:
        if not ctx or not edge_has_annotation(ctx, annotation):
            continue

        values = ctx[ANNOTATIONS][annotation]

        if isinstance(values, str):
            results[values].append(exc.__class__.__name__)

        elif isinstance(values, Iterable):
            for value in values:
                results[value].append(exc.__class__.__name__)

    return dict(results)



# -*- coding: utf-8 -*-

"""Query builder."""

import json
import logging
from typing import Dict, Iterable, List, Mapping, Optional, Set, TextIO, Union

from .exc import QueryMissingNetworksError
from .seeding import Seeding
from ..pipeline import Pipeline
from ...dsl import BaseEntity

__all__ = [
    "Query",
]

logger = logging.getLogger(__name__)


class Query:
    """Represents a query over a network store."""

    def __init__(
        self,
        network_ids: Union[None, int, Iterable[int]] = None,
        seeding: Optional[Seeding] = None,
        pipeline: Optional[Pipeline] = None,
    ) -> None:
        """Build a query.

        :param network_ids: Database network identifiers identifiers
        """
        if not network_ids:
            self.network_ids = []
        elif isinstance(network_ids, int):
            self.network_ids = [network_ids]
        elif isinstance(network_ids, Iterable):
            network_ids = list(network_ids)

            for network_id in network_ids:
                if not isinstance(network_id, int):
                    raise TypeError(network_ids)

            self.network_ids = network_ids
        else:
            raise TypeError(network_ids)

        if seeding is not None and not isinstance(seeding, Seeding):
            raise TypeError("Not a Seeding: {}".format(seeding))
        self.seeding = seeding or Seeding()

        if pipeline is not None and not isinstance(pipeline, Pipeline):
            raise TypeError("Not a pipeline: {}".format(pipeline))
        self.pipeline = pipeline or Pipeline()

    def append_network(self, network_id: int) -> "Query":
        """Add a network to this query.

        :param network_id: The database identifier of the network
        :returns: self for fluid API
        """
        self.network_ids.append(network_id)
        return self

    def append_seeding_induction(self, nodes: Union[BaseEntity, List[BaseEntity], List[Dict]]) -> Seeding:
        """Add a seed induction method.

        :returns: seeding container for fluid API
        """
        return self.seeding.append_induction(nodes)

    def append_seeding_neighbors(self, nodes: Union[BaseEntity, List[BaseEntity], List[Dict]]) -> Seeding:
        """Add a seed by neighbors.

        :returns: seeding container for fluid API
        """
        return self.seeding.append_neighbors(nodes)

    def append_seeding_annotation(self, annotation: str, values: Set[str]) -> Seeding:
        """Add a seed induction method for single annotation's values.

        :param annotation: The annotation to filter by
        :param values: The values of the annotation to keep
        """
        return self.seeding.append_annotation(annotation, values)

    def append_seeding_sample(self, **kwargs) -> Seeding:
        """Add seed induction methods.

        Kwargs can have ``number_edges`` or ``number_seed_nodes``.
        """
        return self.seeding.append_sample(**kwargs)

    def append_pipeline(self, name, *args, **kwargs) -> Pipeline:
        """Add an entry to the pipeline. Defers to :meth:`pybel_tools.pipeline.Pipeline.append`.

        :param name: The name of the function
        :type name: str or types.FunctionType
        :return: This pipeline for fluid query building
        """
        return self.pipeline.append(name, *args, **kwargs)

    def __call__(self, manager):
        """Run this query and returns the resulting BEL graph with :meth:`Query.run`.

        :param pybel.manager.Manager manager: A cache manager
        :rtype: Optional[pybel.BELGraph]
        """
        return self.run(manager)

    def run(self, manager):
        """Run this query and returns the resulting BEL graph.

        :param manager: A cache manager
        :rtype: Optional[pybel.BELGraph]
        """
        universe = self._get_universe(manager)
        graph = self.seeding.run(universe)
        return self.pipeline.run(graph, universe=universe)

    def _get_universe(self, manager):
        if not self.network_ids:
            raise QueryMissingNetworksError("can not run query without network identifiers")

        logger.debug("query universe consists of networks: %s", self.network_ids)

        universe = manager.get_graph_by_ids(self.network_ids)
        logger.debug(
            "query universe has %d nodes/%d edges",
            universe.number_of_nodes(),
            universe.number_of_edges(),
        )

        return universe

    def to_json(self) -> Dict:
        """Return this query as a JSON object."""
        rv = {
            "network_ids": self.network_ids,
        }

        if self.seeding:
            rv["seeding"] = self.seeding.to_json()

        if self.pipeline:
            rv["pipeline"] = self.pipeline.to_json()

        return rv

    def dump(self, file: TextIO, **kwargs) -> None:
        """Dump this query to a file as JSON."""
        json.dump(self.to_json(), file, **kwargs)

    def dumps(self, **kwargs) -> str:
        """Dump this query to a string as JSON."""
        return json.dumps(self.to_json(), **kwargs)

    @staticmethod
    def from_json(data: Mapping) -> "Query":
        """Load a query from a JSON dictionary.

        :param data: A JSON dictionary
        :raises: QueryMissingNetworksError
        """
        network_ids = data.get("network_ids")
        if network_ids is None:
            raise QueryMissingNetworksError('query JSON did not have key "network_ids"')

        seeding_data = data.get("seeding")
        seeding = Seeding.from_json(seeding_data) if seeding_data is not None else None

        pipeline_data = data.get("pipeline")
        pipeline = Pipeline.from_json(pipeline_data) if pipeline_data is not None else None

        return Query(
            network_ids=network_ids,
            seeding=seeding,
            pipeline=pipeline,
        )

    @staticmethod
    def load(file: TextIO) -> "Query":
        """Load a query from a JSON file.

        :raises: QueryMissingNetworksError
        """
        return Query.from_json(json.load(file))

    @staticmethod
    def loads(s: str) -> "Query":
        """Load a query from a JSON string.

        :param s: A stringified JSON query
        :raises: QueryMissingNetworksError
        """
        return Query.from_json(json.loads(s))

    def __str__(self):
        return "Query(networks={}, seeding={}, pipeline={})".format(self.network_ids, self.seeding, self.pipeline)



# -*- coding: utf-8 -*-

"""Functions for expanding a graph based on the upstream/downstream edges."""

from ..utils import expand_by_edge_filter
from ...filters.edge_predicate_builders import (
    build_downstream_edge_predicate,
    build_upstream_edge_predicate,
)
from ...pipeline import uni_in_place_transformation

__all__ = [
    "expand_upstream_causal",
    "expand_downstream_causal",
]


@uni_in_place_transformation
def expand_upstream_causal(universe, graph):
    """Add the upstream causal relations to the given sub-graph.

    :param pybel.BELGraph universe: A BEL graph representing the universe of all knowledge
    :param pybel.BELGraph graph: The target BEL graph to enrich with upstream causal controllers of contained nodes
    """
    expand_by_edge_filter(universe, graph, build_upstream_edge_predicate(graph))


@uni_in_place_transformation
def expand_downstream_causal(universe, graph):
    """Add the downstream causal relations to the given sub-graph.

    :param pybel.BELGraph universe: A BEL graph representing the universe of all knowledge
    :param pybel.BELGraph graph: The target BEL graph to enrich with upstream causal controllers of contained nodes
    """
    expand_by_edge_filter(universe, graph, build_downstream_edge_predicate(graph))



# -*- coding: utf-8 -*-

"""Exceptions for the query builder."""

__all__ = [
    "QueryMissingNetworksError",
    "NodeDegreeIterError",
]


class QueryMissingNetworksError(KeyError):
    """Raised if a query is created from json but doesn't have a listing of network identifiers."""


class NodeDegreeIterError(ValueError):
    """Raised when failing to iterate over node degrees."""



# -*- coding: utf-8 -*-

"""Tests for provenance summary functions."""

import unittest

from pybel.examples import sialic_acid_graph


class TestProvenance(unittest.TestCase):
    """Tests for provenance summary functions."""

    def test_count_citations(self):
        """Test counting citations."""
        count = sialic_acid_graph.number_of_citations()
        self.assertEqual(1, count)



# -*- coding: utf-8 -*-

"""Test parsing metadata from a BEL script."""

import logging
import os
import re
import unittest
from pathlib import Path

from bel_resources.constants import ANNOTATION_URL_FMT, NAMESPACE_URL_FMT

from pybel.exceptions import (
    InvalidMetadataException,
    RedefinedAnnotationError,
    RedefinedNamespaceError,
    VersionFormatWarning,
)
from pybel.parser import MetadataParser
from pybel.resources import HGNC_URL
from pybel.testing.cases import FleetingTemporaryCacheMixin
from pybel.testing.constants import test_an_1, test_ns_1
from pybel.testing.mocks import mock_bel_resources
from tests.constants import (
    HGNC_KEYWORD,
    MESH_DISEASES_KEYWORD,
    MESH_DISEASES_URL,
    help_check_hgnc,
)

logging.getLogger("requests").setLevel(logging.WARNING)

LOCAL_TEST_PATH = os.path.expanduser("~/dev/pybel/src/pybel/testing/resources/belns/hgnc-names.belns")


class TestParseMetadata(FleetingTemporaryCacheMixin):
    def setUp(self):
        super(TestParseMetadata, self).setUp()
        self.parser = MetadataParser(manager=self.manager)

    def _help_test_local_annotation(self, annotation: str) -> None:
        """Check that the annotation is defined locally."""
        self.assertTrue(self.parser.has_annotation(annotation))
        self.assertNotIn(annotation, self.parser.annotation_to_term)
        self.assertFalse(self.parser.has_enumerated_annotation(annotation))
        self.assertNotIn(annotation, self.parser.annotation_to_pattern)
        self.assertFalse(self.parser.has_regex_annotation(annotation))
        self.assertIn(annotation, self.parser.annotation_to_local)
        self.assertTrue(self.parser.has_local_annotation(annotation))

    @mock_bel_resources
    def test_namespace_name_persistience(self, mock_get):
        """Test that a namespace defined by a URL can't be overwritten by a definition by another URL."""
        s = NAMESPACE_URL_FMT.format(HGNC_KEYWORD, HGNC_URL)
        self.parser.parseString(s)
        self.parser.ensure_resources()
        help_check_hgnc(self, self.parser.namespace_to_term_to_encoding)

        s = NAMESPACE_URL_FMT.format(HGNC_KEYWORD, "XXXXX")
        with self.assertRaises(RedefinedNamespaceError):
            self.parser.parseString(s)

        help_check_hgnc(self, self.parser.namespace_to_term_to_encoding)

    @mock_bel_resources
    def test_annotation_name_persistience_1(self, mock_get):
        """Test that an annotation defined by a URL can't be overwritten by a definition by a list."""
        s = ANNOTATION_URL_FMT.format(MESH_DISEASES_KEYWORD, MESH_DISEASES_URL)
        self.parser.parseString(s)
        self.parser.ensure_resources()

        self.assertIn(MESH_DISEASES_KEYWORD, self.parser.annotation_to_term)

        s = 'DEFINE ANNOTATION {} AS LIST {{"A","B","C"}}'.format(MESH_DISEASES_KEYWORD)
        with self.assertRaises(RedefinedAnnotationError):
            self.parser.parseString(s)

        self.assertIn(MESH_DISEASES_KEYWORD, self.parser.annotation_to_term)
        self.assertNotIn("A", self.parser.annotation_to_term[MESH_DISEASES_KEYWORD])
        self.assertIn(
            "46, XX Disorders of Sex Development",
            self.parser.annotation_to_term[MESH_DISEASES_KEYWORD],
        )

    def test_annotation_name_persistience_2(self):
        """Tests that an annotation defined by a list can't be overwritten by a definition by URL"""
        s = 'DEFINE ANNOTATION TextLocation AS LIST {"Abstract","Results","Legend","Review"}'
        self.parser.parseString(s)
        self._help_test_local_annotation("TextLocation")

        s = ANNOTATION_URL_FMT.format("TextLocation", MESH_DISEASES_URL)
        with self.assertRaises(RedefinedAnnotationError):
            self.parser.parseString(s)

        self._help_test_local_annotation("TextLocation")
        self.assertIn("Abstract", self.parser.annotation_to_local["TextLocation"])

    def test_underscore(self):
        """Tests that an underscore is a valid character in an annotation name"""
        s = 'DEFINE ANNOTATION Text_Location AS LIST {"Abstract","Results","Legend","Review"}'
        self.parser.parseString(s)
        self._help_test_local_annotation("Text_Location")

    @mock_bel_resources
    def test_control_compound(self, mock_get):
        text_location = "TextLocation"
        lines = [
            ANNOTATION_URL_FMT.format(MESH_DISEASES_KEYWORD, MESH_DISEASES_URL),
            NAMESPACE_URL_FMT.format(HGNC_KEYWORD, HGNC_URL),
            'DEFINE ANNOTATION TextLocation AS LIST {"Abstract","Results","Legend","Review"}',
        ]
        self.parser.parse_lines(lines)
        self.parser.ensure_resources()

        self.assertIn(MESH_DISEASES_KEYWORD, self.parser.annotation_to_term)
        self.assertIn(HGNC_KEYWORD, self.parser.namespace_to_term_to_encoding)
        self._help_test_local_annotation(text_location)

    @unittest.skipUnless(os.path.exists(LOCAL_TEST_PATH), "Need local files to test local files")
    def test_squiggly_filepath(self):
        line = NAMESPACE_URL_FMT.format(HGNC_KEYWORD, LOCAL_TEST_PATH)
        self.parser.parseString(line)
        help_check_hgnc(self, self.parser.namespace_to_term_to_encoding)

    def test_document_metadata_exception(self):
        s = 'SET DOCUMENT InvalidKey = "nope"'
        with self.assertRaises(InvalidMetadataException):
            self.parser.parseString(s)

    def test_parse_document(self):
        s = '''SET DOCUMENT Name = "Alzheimer's Disease Model"'''

        self.parser.parseString(s)

        self.assertIn("name", self.parser.document_metadata)
        self.assertEqual("Alzheimer's Disease Model", self.parser.document_metadata["name"])

        # Check nothing bad happens
        # with self.assertLogs('pybel', level='WARNING'):
        self.parser.parseString(s)

    @mock_bel_resources
    def test_parse_namespace_url_file(self, mock):
        """Tests parsing a namespace by file URL"""
        s = NAMESPACE_URL_FMT.format("TESTNS1", test_ns_1)
        self.parser.parseString(s)
        self.parser.ensure_resources()

        expected_values = {
            "TestValue1": {"O"},
            "TestValue2": {"O"},
            "TestValue3": {"O"},
            "TestValue4": {"O"},
            "TestValue5": {"O"},
        }

        self.assertIn("TESTNS1", self.parser.namespace_to_term_to_encoding)

        for k, values in expected_values.items():
            k = (None, k)
            self.assertIn(k, self.parser.namespace_to_term_to_encoding["TESTNS1"])
            self.assertEqual(
                set(values),
                set(self.parser.namespace_to_term_to_encoding["TESTNS1"][k]),
            )

    def test_parse_annotation_url_file(self):
        """Tests parsing an annotation by file URL"""
        keyword = "TESTAN1"
        url = Path(test_an_1).as_uri()
        line = ANNOTATION_URL_FMT.format(keyword, url)
        self.parser.parseString(line)
        self.parser.ensure_resources()

        expected_values = {
            "TestAnnot1": "O",
            "TestAnnot2": "O",
            "TestAnnot3": "O",
            "TestAnnot4": "O",
            "TestAnnot5": "O",
        }

        annotation = self.parser.manager.get_namespace_by_url(url)
        self.assertIsNotNone(annotation)
        self.assertEqual(set(expected_values), {e.name for e in annotation.entries})

    def test_parse_annotation_pattern(self):
        s = r'DEFINE ANNOTATION Test AS PATTERN "\w+"'
        self.parser.parseString(s)

        self.assertNotIn("Test", self.parser.annotation_to_term)
        self.assertIn("Test", self.parser.annotation_to_pattern)
        self.assertEqual(re.compile(r"\w+"), self.parser.annotation_to_pattern["Test"])

    def test_define_namespace_regex(self):
        for s, namespace, regex in [
            (
                'DEFINE NAMESPACE dbSNP AS PATTERN "rs[0-9]*"',
                "dbSNP",
                re.compile(r"rs[0-9]*"),
            ),
            ('DEFINE NAMESPACE ec-code AS PATTERN ".*"', "ec-code", re.compile(r".*")),
        ]:
            with self.subTest(namespace=namespace):
                self.parser.parseString(s)
                self.assertNotIn(namespace, self.parser.namespace_to_term_to_encoding)
                self.assertIn(namespace, self.parser.namespace_to_pattern)
                self.assertEqual(regex, self.parser.namespace_to_pattern[namespace])

    def test_not_semantic_version(self):
        s = 'SET DOCUMENT Version = "1.0"'
        with self.assertRaises(VersionFormatWarning):
            self.parser.parseString(s)



# -*- coding: utf-8 -*-

"""Test for export functions in PyBEL-Jupyter."""

import unittest

from pybel.examples import sialic_acid_graph
from pybel.io.jupyter import to_html, to_jupyter_str


class TestHTML(unittest.TestCase):
    """Text HTML functions."""

    def test_to_html(self):
        """Test export to HTML."""
        html = to_html(sialic_acid_graph)
        self.assertIsNotNone(html)

    def test_to_jupyter(self):
        """Test export to JavaScript for Jupyter."""
        javascript = to_jupyter_str(sialic_acid_graph)
        self.assertIsNotNone(javascript)



# -*- coding: utf-8 -*-

"""This module contains simple wrappers around node DSL functions for common namespaces."""

from typing import Optional

from .node_classes import Abundance, MicroRna, Protein

__all__ = [
    "chebi",
    "hgnc",
    "mirbase",
]


def chebi(*, name: Optional[str] = None, identifier: Optional[str] = None) -> Abundance:
    """Build a ChEBI abundance node."""
    return Abundance(namespace="CHEBI", name=name, identifier=identifier)


def hgnc(*, name: Optional[str] = None, identifier: Optional[str] = None) -> Protein:
    """Build an HGNC protein node."""
    return Protein(namespace="HGNC", name=name, identifier=identifier)


def mirbase(*, name: Optional[str] = None, identifier: Optional[str] = None) -> MicroRna:
    """Build an miRBase micro-rna node."""
    return MicroRna(namespace="MIRBASE", name=name, identifier=identifier)



# -*- coding: utf-8 -*-

"""Transport functions for `Fraunhofer's OrientDB <http://graphstore.scai.fraunhofer.de>`_.

`Fraunhofer <https://www.scai.fraunhofer.de/en/business-research-areas/bioinformatics.html>`_ hosts
an instance of `OrientDB <https://orientdb.com/>`_ that contains BEL in a schema similar to
:mod:`pybel.io.umbrella_nodelink`. However, they include custom relations that do not come
from a controlled vocabulary, and have not made the schema, ETL scripts, or documentation available.

Unlike BioDati and BEL Commons, the Fraunhofer OrientDB does not allow for uploads, so only
a single function :func:`pybel.from_fraunhofer_orientdb` is provided by PyBEL.
"""

import logging
from typing import Any, Iterable, Mapping, Optional
from urllib.parse import quote_plus

import requests
from pyparsing import ParseException

from .. import constants as pc
from ..parser import BELParser
from ..struct import BELGraph

__all__ = [
    "from_fraunhofer_orientdb",
]

logger = logging.getLogger(__name__)


def from_fraunhofer_orientdb(  # noqa:S107
    database: str = "covid",
    user: str = "covid_user",
    password: str = "covid",
    query: Optional[str] = None,
) -> BELGraph:
    """Get a BEL graph from the Fraunhofer OrientDB.

    :param database: The OrientDB database to connect to
    :param user: The user to connect to OrientDB
    :param password: The password to connect to OrientDB
    :param query: The query to run. Defaults to the URL encoded version of ``select from E``,
     where ``E`` is all edges in the OrientDB edge database. Likely does not need to be changed,
     except in the case of selecting specific subsets of edges. Make sure you URL encode it
     properly, because OrientDB's RESTful API puts it in the URL's path.

    By default, this function connects to the ``covid`` database, that corresponds to the
    COVID-19 Knowledge Graph [0]_. If other databases in the Fraunhofer OrientDB are
    published and demo username/password combinations are given, the following table will
    be updated.

    +----------+------------+----------+
    | Database | Username   | Password |
    +==========+============+==========+
    | covid    | covid_user | covid    |
    +----------+------------+----------+

    The ``covid`` database can be downloaded and converted to a BEL graph like this:

    .. code-block:: python

        import pybel
        graph = pybel.from_fraunhofer_orientdb(
            database='covid',
            user='covid_user',
            password='covid',
        )
        graph.summarize()

    However, because the source BEL scripts for the COVID-19 Knowledge Graph are available on
    `GitHub <https://github.com/covid19kg/covid19kg>`_ and the authors pre-enabled it for PyBEL, it can
    be downloaded with ``pip install git+https://github.com/covid19kg/covid19kg.git`` and used
    with the following python code:

    .. code-block:: python

       import covid19kg
       graph = covid19kg.get_graph()
       graph.summarize()

    .. warning::

        It was initially planned to handle some of the non-standard relationships listed in the
        Fraunhofer OrientDB's `schema <http://graphstore.scai.fraunhofer.de/studio/index.html#/database/covid/schema>`_
        in their OrientDB Studio instance, but none of them actually appear in the only network that is accessible.
        If this changes, please leave an issue at https://github.com/pybel/pybel/issues so it can be addressed.

    .. [0] Domingo-Fernández, D., *et al.* (2020). `COVID-19 Knowledge Graph: a computable, multi-modal,
           cause-and-effect knowledge model of COVID-19 pathophysiology
           <https://doi.org/10.1101/2020.04.14.040667>`_. *bioRxiv* 2020.04.14.040667.
    """
    graph = BELGraph(name="Fraunhofer OrientDB: {}".format(database))
    parser = BELParser(graph, skip_validation=True)
    results = _request_graphstore(database, user, password, select_query_template=query)
    for result in results:
        _parse_result(parser, result)
    return graph


def _parse_result(parser: BELParser, result: Mapping[str, Any]) -> None:
    citation_db, citation_id = pc.CITATION_TYPE_PUBMED, result.get("pmid")
    if citation_id is None:
        citation_db, citation_id = pc.CITATION_TYPE_PMC, result.get("pmc")
    if citation_id is None:
        if "citation" in result:
            logger.warning(
                "incorrect citation information for %s: %s",
                result["@rid"],
                result["citation"],
            )
        else:
            logger.debug("no citation information for %s", result["@rid"])
        return

    parser.control_parser.clear()
    parser.control_parser.citation_db = citation_db
    parser.control_parser.citation_db_id = citation_id
    parser.control_parser.evidence = result["evidence"]
    parser.control_parser.annotations.update(result["annotation"])

    source = result["in"]["bel"]
    relation = result["@class"]
    relation = RELATION_MAP.get(relation, relation)
    target = result["out"]["bel"]
    statement = " ".join([source, relation, target])

    try:
        parser.parseString(statement)
    except ParseException:
        logger.warning("could not parse %s", statement)


RELATION_MAP = {
    "causes_no_change": pc.CAUSES_NO_CHANGE,
    "positive_correlation": pc.POSITIVE_CORRELATION,
    "negative_correlation": pc.NEGATIVE_CORRELATION,
    "is_a": pc.IS_A,
    "has_member": "hasMember",
    "has_members": "hasMembers",
    "has_component": "hasComponent",
    "has_components": "hasComponents",
}


def _request_graphstore(
    database: str,
    user: str,
    password: str,
    count_query: Optional[str] = None,
    select_query_template: Optional[str] = None,
    page_size: int = 500,
    base: str = "http://graphstore.scai.fraunhofer.de/query",
) -> Iterable[Mapping[str, Any]]:
    """Make an API call to the OrientDB."""
    if count_query is None:
        count_query = "select count(@rid) from E"
    count_query = quote_plus(count_query)
    count_url = "{base}/{database}/sql/{count_query}".format(base=base, database=database, count_query=count_query)
    count_res = requests.get(count_url, auth=(user, password))
    count = count_res.json()["result"][0]["count"]
    logging.debug("fraunhofer orientdb has %d edges", count)

    if select_query_template is None:
        select_query_template = "select from E order by @rid limit {limit} offset {offset}"

    offsets = count // page_size
    for offset in range(offsets + 1):
        select_query = select_query_template.format(limit=page_size, offset=offset * page_size)
        logger.debug("query: %s", select_query)
        select_query = quote_plus(select_query)
        select_url = "{base}/{database}/sql/{select_query}/{page_size}/*:1".format(
            base=base,
            database=database,
            select_query=select_query,
            page_size=page_size,
        )
        res = requests.get(select_url, auth=(user, password))
        res_json = res.json()
        result = res_json["result"]
        yield from result



# -*- coding: utf-8 -*-

"""Test cases for PyBEL-CX."""

import json
import unittest
from typing import Mapping, Tuple

from pybel import BaseEntity, BELGraph
from pybel.constants import (
    ANNOTATIONS,
    CITATION,
    EVIDENCE,
    IDENTIFIER,
    NAMESPACE,
    RELATION,
    SOURCE_MODIFIER,
    TARGET_MODIFIER,
)
from pybel.typing import EdgeData

__all__ = [
    "TestCase",
]


def _edge_to_tuple(u: BaseEntity, v: BaseEntity, edge_data: EdgeData):
    """Convert an edge to tuple.

    :return: A tuple that can be hashed representing this edge. Makes no promises to its structure.
    """
    citation = edge_data.get(CITATION)
    if citation is None:
        citation_hashable = None
    else:
        citation_hashable = (citation[NAMESPACE], citation[IDENTIFIER])

    evidence_hashable = edge_data.get(EVIDENCE)

    annotations = edge_data.get(ANNOTATIONS)
    if annotations is None:
        annotations_hashable = None
    else:
        annotations_hashable = tuple((key, tuple(sorted(values))) for key, values in sorted(annotations.items()))

    source_modifier = edge_data.get(SOURCE_MODIFIER)
    if source_modifier is None:
        subject_hashable = None
    else:
        subject_hashable = json.dumps(source_modifier, ensure_ascii=True, sort_keys=True, indent=0)

    target_modifier = edge_data.get(TARGET_MODIFIER)
    if target_modifier is None:
        object_hashable = None
    else:
        object_hashable = json.dumps(target_modifier, ensure_ascii=True, sort_keys=True, indent=0)

    return (
        u,
        v,
        edge_data[RELATION],
        citation_hashable,
        evidence_hashable,
        annotations_hashable,
        subject_hashable,
        object_hashable,
    )


def _hash_edge(u: BaseEntity, v: BaseEntity, data: EdgeData) -> int:
    """Convert an edge tuple to a hash.

    :return: A hashed version of the edge tuple using md5 hash of the binary pickle dump of u, v, and the json dump of d
    """
    return hash(_edge_to_tuple(u, v, data))


def _get_edge_dict(
    graph: BELGraph,
) -> Mapping[int, Tuple[BaseEntity, BaseEntity, EdgeData]]:
    return {_hash_edge(u, v, data): (u, v, data) for u, v, k, data in graph.edges(keys=True, data=True)}


class TestCase(unittest.TestCase):
    """Extension to base :class:`unittest.TestCase` with :class:`pybel.BELGraph` comparison."""

    def assert_graph_equal(self, g1: BELGraph, g2: BELGraph) -> None:
        """Assert two BEL graphs are the same."""
        self.assertEqual(g1.graph, g2.graph, msg="Metadata were not the same")

        # self.assertEqual(g1.number_of_nodes(), g2.number_of_nodes())
        self.assertEqual(set(g1), set(g2), msg="Nodes were not the same")

        # self.assertEqual(g1.number_of_edges(), g2.number_of_edges())
        self.assertEqual(set(g1.edges()), set(g2.edges()))

        g1_edge_hashes = _get_edge_dict(g1)
        g2_edge_hashes = _get_edge_dict(g2)

        g1k = g1_edge_hashes.keys()
        g2k = g2_edge_hashes.keys()

        g1ng2 = g1k - g2k
        g2ng1 = g2k - g1k

        if g1ng2 and not g2ng1:
            for k in g1ng2:
                print(k[:6], g1_edge_hashes[k])
            self.fail()

        elif not g1ng2 and g2ng1:
            for k in g2ng1:
                print(k[:6], g2_edge_hashes[k])
            self.fail()

        elif g1ng2 and g2ng1:
            print("in g1 but not g2:")
            for k in g1ng2:
                print(k[:6], g1_edge_hashes[k])

            print("in g2 but not g1:")
            for k in g2ng1:
                print(k[:6], g2_edge_hashes[k])
            self.fail()



# -*- coding: utf-8 -*-

"""Tests for the query builder."""

import logging
import unittest

from pybel import BELGraph
from pybel.dsl import Protein
from pybel.examples.egf_example import egf_graph
from pybel.struct.query import Seeding
from pybel.testing.generate import generate_random_graph
from pybel.testing.utils import n

log = logging.getLogger(__name__)


class TestSeedingConstructor(unittest.TestCase):
    def test_none(self):
        seeding = Seeding()
        self.assertEqual(0, len(seeding))
        self.assertEqual("[]", seeding.dumps())

    def test_append_sample(self):
        seeding = Seeding()
        seeding.append_sample()
        self.assertEqual(1, len(seeding))

        s = seeding.dumps()
        self.assertIsInstance(s, str)

    def test_no_seeding(self):
        graph = egf_graph.copy()

        seeding = Seeding()
        result = seeding.run(graph)

        self.assertEqual(graph.number_of_nodes(), result.number_of_nodes())
        self.assertEqual(graph.number_of_edges(), result.number_of_edges())

    def test_seed_by_neighbor(self):
        graph = BELGraph()
        a, b, c, d = (Protein(namespace=n(), name=str(i)) for i in range(4))
        graph.add_increases(a, b, citation=n(), evidence=n())
        graph.add_increases(b, c, citation=n(), evidence=n())
        graph.add_increases(c, d, citation=n(), evidence=n())

        seeding = Seeding()
        seeding.append_neighbors(b)
        result = seeding.run(graph)

        self.assertIsInstance(result, BELGraph)
        # test nodes
        self.assertIn(a, result)
        self.assertIn(b, result)
        self.assertIn(c, result)
        self.assertNotIn(d, result)
        # test edges
        self.assertIn(b, result[a])
        self.assertIn(c, result[b])
        self.assertNotIn(d, result[c])

    def test_seed_by_neighbors(self):
        graph = BELGraph()
        a, b, c, d, e = (Protein(namespace=n(), name=str(i)) for i in range(5))
        graph.add_increases(a, b, citation=n(), evidence=n())
        graph.add_increases(b, c, citation=n(), evidence=n())
        graph.add_increases(c, d, citation=n(), evidence=n())
        graph.add_increases(d, e, citation=n(), evidence=n())

        seeding = Seeding()
        seeding.append_neighbors([b, c])
        result = seeding.run(graph)

        self.assertIsInstance(result, BELGraph)
        # test nodes
        self.assertIn(a, result)
        self.assertIn(b, result)
        self.assertIn(c, result)
        self.assertIn(d, result)
        self.assertNotIn(e, result)
        # test edges
        self.assertIn(b, result[a])
        self.assertIn(c, result[b])
        self.assertIn(d, result[c])
        self.assertNotIn(e, result[d])

    def test_random_sample(self):
        graph = generate_random_graph(50, 1000)

        seeding = Seeding()
        seeding.append_sample(number_edges=10)
        seeding.append_sample(number_edges=10)
        result = seeding.run(graph)

        # TODO this will fail randomly some times lol, so make allowed to be sort of wrong
        self.assertIn(result.number_of_edges(), {18, 19, 20})



# -*- coding: utf-8 -*-

"""Tests for PyBEL induction functions."""

import string
import unittest

from pybel import BELGraph
from pybel.constants import (
    CITATION_AUTHORS,
    CITATION_TYPE_PUBMED,
    IDENTIFIER,
    NAMESPACE,
)
from pybel.dsl import BaseEntity, gene, protein, rna
from pybel.struct.mutation.expansion import expand_upstream_causal
from pybel.struct.mutation.induction import get_subgraph_by_annotation_value
from pybel.struct.mutation.induction.citation import (
    get_subgraph_by_authors,
    get_subgraph_by_pubmed,
)
from pybel.struct.mutation.induction.paths import (
    get_nodes_in_all_shortest_paths,
    get_subgraph_by_all_shortest_paths,
)
from pybel.struct.mutation.induction.upstream import get_upstream_causal_subgraph
from pybel.struct.mutation.induction.utils import get_subgraph_by_induction
from pybel.testing.utils import n

trem2_gene = gene(namespace="HGNC", name="TREM2")
trem2_rna = rna(namespace="HGNC", name="TREM2")
trem2_protein = protein(namespace="HGNC", name="TREM2")


class TestGraphMixin(unittest.TestCase):
    """A mixin to enable testing nodes and edge membership in the graph."""

    def assert_in_edge(self, source, target, graph):
        """Assert the edge is in the graph.

        :param source:
        :param target:
        :type graph: pybel.BELGraph
        :rtype: bool
        """
        self.assertIn(target, graph[source])

    def assert_all_nodes_are_base_entities(self, graph):
        """Assert that all nodes are base entities."""
        for node in graph:
            self.assertIsInstance(node, BaseEntity)


class TestInduction(TestGraphMixin):
    """Test induction functions."""

    def test_get_subgraph_by_induction(self):
        """Test get_subgraph_by_induction."""
        graph = BELGraph()
        keyword, url = n(), n()
        graph.namespace_url[keyword] = url
        a, b, c, d = [protein(namespace="test", name=str(i)) for i in range(4)]
        graph.add_directly_increases(a, b, citation=n(), evidence=n())
        graph.add_directly_increases(b, c, citation=n(), evidence=n())
        graph.add_directly_increases(c, d, citation=n(), evidence=n())
        graph.add_increases(a, d, citation=n(), evidence=n())

        nodes = [b, c]
        subgraph = get_subgraph_by_induction(graph, nodes)

        self.assertIsInstance(subgraph, BELGraph)
        self.assert_all_nodes_are_base_entities(subgraph)
        self.assertNotEqual(
            0,
            len(subgraph.namespace_url),
            msg="improperly found metadata: {}".format(subgraph.graph),
        )
        self.assertIn(keyword, subgraph.namespace_url)
        self.assertEqual(url, subgraph.namespace_url[keyword])

        self.assertNotIn(a, subgraph)
        self.assertIn(b, subgraph)
        self.assertIn(c, subgraph)
        self.assertNotIn(d, subgraph)

    def test_get_subgraph_by_all_shortest_paths(self):
        """Test get_subgraph_by_all_shortest_paths."""
        graph = BELGraph()
        keyword, url = n(), n()
        graph.namespace_url[keyword] = url
        a, b, c, d, e, f = [protein(namespace="test", name=n()) for _ in range(6)]
        graph.add_increases(a, b, citation=n(), evidence=n())
        graph.add_increases(a, c, citation=n(), evidence=n())
        graph.add_increases(b, d, citation=n(), evidence=n())
        graph.add_increases(c, d, citation=n(), evidence=n())
        graph.add_increases(a, e, citation=n(), evidence=n())
        graph.add_increases(e, f, citation=n(), evidence=n())
        graph.add_increases(f, d, citation=n(), evidence=n())

        query_nodes = [a, d]
        shortest_paths_nodes = get_nodes_in_all_shortest_paths(graph, query_nodes)

        self.assertIn(a, shortest_paths_nodes)
        self.assertIn(b, shortest_paths_nodes)
        self.assertIn(c, shortest_paths_nodes)
        self.assertIn(d, shortest_paths_nodes)

        subgraph = get_subgraph_by_all_shortest_paths(graph, query_nodes)
        self.assert_all_nodes_are_base_entities(subgraph)
        self.assertIsInstance(subgraph, BELGraph)
        self.assertIn(keyword, subgraph.namespace_url)
        self.assertEqual(url, subgraph.namespace_url[keyword])

        self.assertIn(a, subgraph)
        self.assertIn(b, subgraph)
        self.assertIn(c, subgraph)
        self.assertIn(d, subgraph)
        self.assertNotIn(e, subgraph)
        self.assertNotIn(f, subgraph)

    def test_get_upstream_causal_subgraph(self):
        """Test get_upstream_causal_subgraph."""
        a, b, c, d, e, f = [protein(namespace="test", name=n()) for _ in range(6)]

        universe = BELGraph()
        universe.namespace_pattern["test"] = "test-url"
        universe.add_increases(a, b, citation=n(), evidence=n())
        universe.add_increases(b, c, citation=n(), evidence=n())
        universe.add_association(d, a, citation=n(), evidence=n())
        universe.add_increases(e, a, citation=n(), evidence=n())
        universe.add_decreases(f, b, citation=n(), evidence=n())

        subgraph = get_upstream_causal_subgraph(universe, [a, b])

        self.assertIsInstance(subgraph, BELGraph)
        self.assert_all_nodes_are_base_entities(subgraph)

        self.assertIn("test", subgraph.namespace_pattern)
        self.assertEqual("test-url", subgraph.namespace_pattern["test"])

        self.assertIn(a, subgraph)
        self.assertIn(b, subgraph)
        self.assertNotIn(c, subgraph)
        self.assertNotIn(d, subgraph)
        self.assertIn(e, subgraph)
        self.assertIn(f, subgraph)
        self.assertEqual(4, subgraph.number_of_nodes())

        self.assert_in_edge(e, a, subgraph)
        self.assert_in_edge(a, b, subgraph)
        self.assert_in_edge(f, b, subgraph)
        self.assertEqual(3, subgraph.number_of_edges())

    def test_expand_upstream_causal_subgraph(self):
        """Test expanding on the upstream causal subgraph."""
        a, b, c, d, e, f = [protein(namespace="test", name=i) for i in string.ascii_lowercase[:6]]

        universe = BELGraph()
        universe.add_increases(a, b, citation=n(), evidence=n())
        universe.add_increases(b, c, citation=n(), evidence=n())
        universe.add_association(d, a, citation=n(), evidence=n())
        universe.add_increases(e, a, citation=n(), evidence=n())
        universe.add_decreases(f, b, citation=n(), evidence=n())

        subgraph = BELGraph()
        subgraph.add_increases(a, b, citation=n(), evidence=n())

        expand_upstream_causal(universe, subgraph)

        self.assertIsInstance(subgraph, BELGraph)
        self.assert_all_nodes_are_base_entities(subgraph)

        self.assertIn(a, subgraph)
        self.assertIn(b, subgraph)
        self.assertNotIn(c, subgraph)
        self.assertNotIn(d, subgraph)
        self.assertIn(e, subgraph)
        self.assertIn(f, subgraph)
        self.assertEqual(4, subgraph.number_of_nodes())

        self.assert_in_edge(e, a, subgraph)
        self.assert_in_edge(a, b, subgraph)
        self.assert_in_edge(f, b, subgraph)
        self.assertEqual(2, len(subgraph[a][b]))
        self.assertEqual(4, subgraph.number_of_edges(), msg="\n".join(map(str, subgraph.edges())))


class TestEdgePredicateBuilders(TestGraphMixin):
    """Tests for edge predicate builders."""

    def test_build_pmid_inclusion_filter(self):
        """Test getting a sub-graph by a single PubMed identifier."""
        a, b, c, d = [protein(namespace="test", name=n()) for _ in range(4)]
        p1, p2, p3, p4 = n(), n(), n(), n()

        graph = BELGraph()
        keyword, url = n(), n()
        graph.namespace_url[keyword] = url
        graph.add_increases(a, b, evidence=n(), citation=p1)
        graph.add_increases(a, b, evidence=n(), citation=p2)
        graph.add_increases(b, c, evidence=n(), citation=p1)
        graph.add_increases(b, c, evidence=n(), citation=p3)
        graph.add_increases(c, d, evidence=n(), citation=p3)

        subgraph = get_subgraph_by_pubmed(graph, p1)

        self.assertIsInstance(subgraph, BELGraph)
        self.assert_all_nodes_are_base_entities(subgraph)

        self.assertIn(keyword, subgraph.namespace_url)
        self.assertEqual(url, subgraph.namespace_url[keyword])

        self.assertIn(a, subgraph)
        self.assertIn(b, subgraph)
        self.assertIn(c, subgraph)
        self.assertNotIn(d, subgraph)

        empty_subgraph = get_subgraph_by_pubmed(graph, p4)
        self.assertIn(keyword, subgraph.namespace_url)
        self.assertEqual(url, subgraph.namespace_url[keyword])
        self.assertEqual(0, empty_subgraph.number_of_nodes())

    def test_build_pmid_set_inclusion_filter(self):
        """Test getting a sub-graph by a set of PubMed identifiers."""
        a, b, c, d, e, f = [protein(namespace="test", name=n()) for _ in range(6)]
        p1, p2, p3, p4, p5, p6 = n(), n(), n(), n(), n(), n()

        graph = BELGraph()
        keyword, url = n(), n()
        graph.namespace_url[keyword] = url
        graph.add_increases(a, b, evidence=n(), citation=p1)
        graph.add_increases(a, b, evidence=n(), citation=p2)
        graph.add_increases(b, c, evidence=n(), citation=p1)
        graph.add_increases(b, c, evidence=n(), citation=p3)
        graph.add_increases(c, d, evidence=n(), citation=p3)
        graph.add_increases(e, f, evidence=n(), citation=p4)

        subgraph = get_subgraph_by_pubmed(graph, [p1, p4])

        self.assertIsInstance(subgraph, BELGraph)
        self.assert_all_nodes_are_base_entities(subgraph)

        self.assertIn(keyword, subgraph.namespace_url)
        self.assertEqual(url, subgraph.namespace_url[keyword])

        self.assertIn(a, subgraph)
        self.assertIn(b, subgraph)
        self.assertIn(c, subgraph)
        self.assertNotIn(d, subgraph)
        self.assertIn(e, subgraph)
        self.assertIn(f, subgraph)

        empty_subgraph = get_subgraph_by_pubmed(graph, [p5, p6])
        self.assertIn(keyword, subgraph.namespace_url)
        self.assertEqual(url, subgraph.namespace_url[keyword])
        self.assertEqual(0, empty_subgraph.number_of_nodes())

    def test_build_author_inclusion_filter(self):
        """Test getting a sub-graph by a single author."""
        a, b, c, d = [protein(namespace="test", name=n()) for _ in range(4)]
        a1, a2, a3, a4, a5 = n(), n(), n(), n(), n()

        c1 = {
            NAMESPACE: CITATION_TYPE_PUBMED,
            IDENTIFIER: n(),
            CITATION_AUTHORS: [a1, a2, a3],
        }
        c2 = {
            NAMESPACE: CITATION_TYPE_PUBMED,
            IDENTIFIER: n(),
            CITATION_AUTHORS: [a1, a4],
        }

        graph = BELGraph()
        keyword, url = n(), n()
        graph.namespace_url[keyword] = url
        graph.add_increases(a, b, evidence=n(), citation=c1)
        graph.add_increases(a, b, evidence=n(), citation=c2)
        graph.add_increases(b, c, evidence=n(), citation=c1)
        graph.add_increases(c, d, evidence=n(), citation=c2)

        subgraph1 = get_subgraph_by_authors(graph, a1)

        self.assertIsInstance(subgraph1, BELGraph)
        self.assert_all_nodes_are_base_entities(subgraph1)

        self.assertIn(keyword, subgraph1.namespace_url)
        self.assertEqual(url, subgraph1.namespace_url[keyword])

        self.assertIn(a, subgraph1)
        self.assertIn(b, subgraph1)
        self.assertIn(c, subgraph1)
        self.assertIn(d, subgraph1)

        subgraph2 = get_subgraph_by_authors(graph, a2)

        self.assertIsInstance(subgraph2, BELGraph)
        self.assert_all_nodes_are_base_entities(subgraph2)

        self.assertIn(keyword, subgraph2.namespace_url)
        self.assertEqual(url, subgraph2.namespace_url[keyword])

        self.assertIn(a, subgraph2)
        self.assertIn(b, subgraph2)
        self.assertIn(c, subgraph2)
        self.assertNotIn(d, subgraph2)

        subgraph3 = get_subgraph_by_authors(graph, a5)

        self.assertIsInstance(subgraph3, BELGraph)
        self.assert_all_nodes_are_base_entities(subgraph3)

        self.assertIn(keyword, subgraph3.namespace_url)
        self.assertEqual(url, subgraph3.namespace_url[keyword])
        self.assertEqual(0, subgraph3.number_of_nodes())

    def test_build_author_set_inclusion_filter(self):
        """Test getting a sub-graph by a set of authors."""
        a, b, c, d = [protein(namespace="test", name=n()) for _ in range(4)]
        a1, a2, a3, a4 = n(), n(), n(), n()

        c1 = {
            NAMESPACE: CITATION_TYPE_PUBMED,
            IDENTIFIER: n(),
            CITATION_AUTHORS: [a1, a2, a3],
        }
        c2 = {
            NAMESPACE: CITATION_TYPE_PUBMED,
            IDENTIFIER: n(),
            CITATION_AUTHORS: [a1, a4],
        }

        graph = BELGraph()
        keyword, url = n(), n()
        graph.namespace_url[keyword] = url
        graph.add_increases(a, b, evidence=n(), citation=c1)
        graph.add_increases(a, b, evidence=n(), citation=c2)
        graph.add_increases(b, c, evidence=n(), citation=c1)
        graph.add_increases(c, d, evidence=n(), citation=c2)

        subgraph1 = get_subgraph_by_authors(graph, [a1, a2])

        self.assertIsInstance(subgraph1, BELGraph)
        self.assert_all_nodes_are_base_entities(subgraph1)

        self.assertIn(keyword, subgraph1.namespace_url)
        self.assertEqual(url, subgraph1.namespace_url[keyword])

        self.assertIn(a, subgraph1)
        self.assertIn(b, subgraph1)
        self.assertIn(c, subgraph1)
        self.assertIn(d, subgraph1)


class TestEdgeInduction(unittest.TestCase):
    """Test induction over edges."""

    def test_get_subgraph_by_annotation_value(self):
        """Test getting a subgraph by a single annotation value."""
        graph = BELGraph()
        graph.annotation_url["Subgraph"] = n()
        a, b, c, d = [protein(namespace="test", name=n()) for _ in range(4)]

        k1 = graph.add_increases(a, b, citation=n(), evidence=n(), annotations={"Subgraph": {"A"}})

        k2 = graph.add_increases(a, b, citation=n(), evidence=n(), annotations={"Subgraph": {"B"}})

        k3 = graph.add_increases(a, b, citation=n(), evidence=n(), annotations={"Subgraph": {"A", "C", "D"}})

        subgraph = get_subgraph_by_annotation_value(graph, "Subgraph", "A")
        self.assertIsInstance(subgraph, BELGraph)

        self.assertIn(a, subgraph)
        self.assertIn(b, subgraph)
        self.assertIn(b, subgraph[a])
        self.assertIn(k1, subgraph[a][b])
        self.assertNotIn(k2, subgraph[a][b])
        self.assertIn(k3, subgraph[a][b])

    def test_get_subgraph_by_annotation_values(self):
        """Test getting a subgraph by multiple annotation value."""
        graph = BELGraph()
        graph.annotation_list["Subgraph"] = set("ABCDE")

        a, b, c, d = [protein(namespace="test", name=n()) for _ in range(4)]

        k1 = graph.add_increases(a, b, citation=n(), evidence=n(), annotations={"Subgraph": {"A"}})

        k2 = graph.add_increases(a, b, citation=n(), evidence=n(), annotations={"Subgraph": {"B"}})

        k3 = graph.add_increases(a, b, citation=n(), evidence=n(), annotations={"Subgraph": {"A", "C", "D"}})

        k4 = graph.add_increases(a, b, citation=n(), evidence=n(), annotations={"Subgraph": {"C", "D"}})

        subgraph = get_subgraph_by_annotation_value(graph, "Subgraph", {"A", "C"})
        self.assertIsInstance(subgraph, BELGraph)

        self.assertIn(a, subgraph)
        self.assertIn(b, subgraph)
        self.assertIn(b, subgraph[a])
        self.assertIn(k1, subgraph[a][b])
        self.assertNotIn(k2, subgraph[a][b])
        self.assertIn(k3, subgraph[a][b])
        self.assertIn(k4, subgraph[a][b])



# -*- coding: utf-8 -*-

"""Transport functions for `BEL Commons <https://github.com/bel-commons/bel-commons>`_.

BEL Commons is a free, open-source platform for hosting BEL content. Because it was originally
developed and published in an academic capacity at Fraunhofer SCAI, a public instance can be
found at https://bel-commons-dev.scai.fraunhofer.de. However, this instance is only supported
out of posterity and will not be updated. If you would like to host your own instance of
BEL Commons, there are instructions on its GitHub page.
"""

import logging
from typing import Optional

import pystow
import requests

from .nodelink import from_nodelink, to_nodelink
from ..struct.graph import BELGraph
from ..version import get_version

__all__ = [
    "to_bel_commons",
    "from_bel_commons",
]

logger = logging.getLogger(__name__)

RECIEVE_ENDPOINT = "/api/receive/"
GET_ENDPOINT = "/api/network/{}/export/nodelink"


def _get_host() -> Optional[str]:
    """Find the host with :func:`pystow.get_config`.

    Has two possibilities:

    1. The PyBEL config entry ``PYBEL_REMOTE_HOST``, loaded in :mod:`pybel.constants`
    2. The environment variable ``PYBEL_REMOTE_HOST``
    """
    return pystow.get_config("pybel", "remote_host")


def _get_user() -> Optional[str]:
    return pystow.get_config("pybel", "remote_user")


def _get_password() -> Optional[str]:
    return pystow.get_config("pybel", "remote_password")


def to_bel_commons(
    graph: BELGraph,
    host: Optional[str] = None,
    user: Optional[str] = None,
    password: Optional[str] = None,
    public: bool = True,
) -> requests.Response:
    """Send a graph to the receiver service and returns the :mod:`requests` response object.

    :param graph: A BEL graph
    :param host: The location of the BEL Commons server. Alternatively, looks up in PyBEL config with
     ``PYBEL_REMOTE_HOST`` or the environment as ``PYBEL_REMOTE_HOST``.
    :param user: Username for BEL Commons. Alternatively, looks up in PyBEL config with
     ``PYBEL_REMOTE_USER`` or the environment as ``PYBEL_REMOTE_USER``
    :param password: Password for BEL Commons. Alternatively, looks up in PyBEL config with
     ``PYBEL_REMOTE_PASSWORD`` or the environment as ``PYBEL_REMOTE_PASSWORD``
    :param public: Should the network be made public?
    :return: The response object from :mod:`requests`
    """
    if host is None:
        host = _get_host()
        logger.debug("using host: %s", host)

    if user is None:
        user = _get_user()

        if user is None:
            raise ValueError("no user found")

    if password is None:
        password = _get_password()

        if password is None:
            raise ValueError("no password found")

    url = host.rstrip("/") + RECIEVE_ENDPOINT

    response = requests.post(
        url,
        json=to_nodelink(graph),
        headers={
            "content-type": "application/json",
            "User-Agent": "PyBEL v{}".format(get_version()),
            "bel-commons-public": "true" if public else "false",
        },
        auth=(user, password),
    )
    logger.debug("received response: %s", response)

    return response


def from_bel_commons(network_id: int, host: Optional[str] = None) -> BELGraph:
    """Retrieve a public network from BEL Commons.

    In the future, this function may be extended to support authentication.

    :param network_id: The BEL Commons network identifier
    :param host: The location of the BEL Commons server. Alternatively, looks up in PyBEL config with
     ``PYBEL_REMOTE_HOST`` or the environment as ``PYBEL_REMOTE_HOST``.
    :raises: ValueError if host configuration can not be found
    """
    if host is None:
        host = _get_host()
    if host is None:
        raise ValueError("host not specified in arguments, PyBEL configuration, or environment.")

    url = host + GET_ENDPOINT.format(network_id)
    res = requests.get(url)
    graph_json = res.json()
    graph = from_nodelink(graph_json)
    return graph



# -*- coding: utf-8 -*-

"""An exporter for signaling pathway impact analysis (SPIA) described by [Tarca2009]_.

.. [Tarca2009] Tarca, A. L., *et al* (2009). `A novel signaling pathway impact analysis
               <https://doi.org/10.1093/bioinformatics/btn577>`_. Bioinformatics, 25(1), 75–82.

.. seealso:: https://bioconductor.org/packages/release/bioc/html/SPIA.html
"""

import itertools as itt
import os
from collections import OrderedDict
from typing import Dict, Mapping, Set

import pandas as pd

from ..constants import (
    ASSOCIATION,
    CAUSAL_DECREASE_RELATIONS,
    CAUSAL_INCREASE_RELATIONS,
    RELATION,
)
from ..dsl import CentralDogma, Gene, ListAbundance, ProteinModification, Rna
from ..language import pmod_mappings
from ..struct import BELGraph
from ..typing import EdgeData

__all__ = [
    "to_spia_dfs",
    "to_spia_excel",
    "to_spia_tsvs",
]

SPIADataFrames = Mapping[str, pd.DataFrame]

KEGG_RELATIONS = [
    "activation",
    "compound",
    "binding/association",
    "expression",
    "inhibition",
    "activation_phosphorylation",
    "phosphorylation",
    "inhibition_phosphorylation",
    "inhibition_dephosphorylation",
    "dissociation",
    "dephosphorylation",
    "activation_dephosphorylation",
    "state change",
    "activation_indirect effect",
    "inhibition_ubiquination",
    "ubiquination",
    "expression_indirect effect",
    "inhibition_indirect effect",
    "repression",
    "dissociation_phosphorylation",
    "indirect effect_phosphorylation",
    "activation_binding/association",
    "indirect effect",
    "activation_compound",
    "activation_ubiquination",
]


def to_spia_excel(graph: BELGraph, path: str) -> None:
    """Write the BEL graph as an SPIA-formatted excel sheet at the given path."""
    x = to_spia_dfs(graph)
    spia_matrices_to_excel(x, path)


def to_spia_tsvs(graph: BELGraph, directory: str) -> None:
    """Write the BEL graph as a set of SPIA-formatted TSV files in a given directory."""
    x = to_spia_dfs(graph)
    spia_matrices_to_tsvs(x, directory)


def to_spia_dfs(graph: BELGraph) -> SPIADataFrames:
    """Create an excel sheet ready to be used in SPIA software.

    :param graph: BELGraph
    :return: dictionary with matrices
    """
    index_nodes = get_matrix_index(graph)
    spia_matrices = build_spia_matrices(index_nodes)

    for u, v, edge_data in graph.edges(data=True):
        # Both nodes are CentralDogma abundances
        if isinstance(u, CentralDogma) and isinstance(v, CentralDogma):
            # Update matrix dict
            update_spia_matrices(spia_matrices, u, v, edge_data)

        # Subject is CentralDogmaAbundance and node is ListAbundance
        elif isinstance(u, CentralDogma) and isinstance(v, ListAbundance):
            # Add a relationship from subject to each of the members in the object
            for node in v.members:
                # Skip if the member is not in CentralDogma
                if not isinstance(node, CentralDogma):
                    continue

                update_spia_matrices(spia_matrices, u, node, edge_data)

        # Subject is ListAbundance and node is CentralDogmaAbundance
        elif isinstance(u, ListAbundance) and isinstance(v, CentralDogma):
            # Add a relationship from each of the members of the subject to the object
            for node in u.members:
                # Skip if the member is not in CentralDogma
                if not isinstance(node, CentralDogma):
                    continue

                update_spia_matrices(spia_matrices, node, v, edge_data)

        # Both nodes are ListAbundance
        elif isinstance(u, ListAbundance) and isinstance(v, ListAbundance):
            for sub_member, obj_member in itt.product(u.members, v.members):
                # Update matrix if both are CentralDogma
                if isinstance(sub_member, CentralDogma) and isinstance(obj_member, CentralDogma):
                    update_spia_matrices(spia_matrices, sub_member, obj_member, edge_data)

        # else Not valid edge

    return spia_matrices


def get_matrix_index(graph: BELGraph) -> Set[str]:
    """Return set of HGNC names from Proteins/Rnas/Genes/miRNA, nodes that can be used by SPIA."""
    # TODO: Using HGNC Symbols for now
    return {node.name for node in graph if isinstance(node, CentralDogma) and node.namespace.upper() == "HGNC"}


def build_spia_matrices(nodes: Set[str]) -> Dict[str, pd.DataFrame]:
    """Build an adjacency matrix for each KEGG relationship and return in a dictionary.

    :param nodes: A set of HGNC gene symbols
    :return: Dictionary of adjacency matrix for each relationship
    """
    nodes = list(sorted(nodes))

    # Create sheets of the excel in the given order
    matrices = OrderedDict()
    for relation in KEGG_RELATIONS:
        matrices[relation] = pd.DataFrame(0, index=nodes, columns=nodes)

    return matrices


UB_NAMES = {"Ub"} | {e.name for e in pmod_mappings["Ub"]["xrefs"]}
PH_NAMES = {"Ph"} | {e.name for e in pmod_mappings["Ph"]["xrefs"]}


def update_spia_matrices(
    spia_matrices: Dict[str, pd.DataFrame],
    u: CentralDogma,
    v: CentralDogma,
    edge_data: EdgeData,
) -> None:
    """Populate the adjacency matrix."""
    if u.namespace.lower() != "hgnc" or v.namespace.lower() != "hgnc":
        return

    u_name = u.name
    v_name = v.name
    relation = edge_data[RELATION]

    if relation in CAUSAL_INCREASE_RELATIONS:
        # If it has pmod check which one and add it to the corresponding matrix
        if v.variants and any(isinstance(variant, ProteinModification) for variant in v.variants):
            for variant in v.variants:
                if not isinstance(variant, ProteinModification):
                    continue
                elif variant.entity.name in UB_NAMES:
                    spia_matrices["activation_ubiquination"][u_name][v_name] = 1
                elif variant.entity.name in PH_NAMES:
                    spia_matrices["activation_phosphorylation"][u_name][v_name] = 1
        elif isinstance(v, (Gene, Rna)):  # Normal increase, add activation
            spia_matrices["expression"][u_name][v_name] = 1
        else:
            spia_matrices["activation"][u_name][v_name] = 1

    elif relation in CAUSAL_DECREASE_RELATIONS:
        # If it has pmod check which one and add it to the corresponding matrix
        if v.variants and any(isinstance(variant, ProteinModification) for variant in v.variants):
            for variant in v.variants:
                if not isinstance(variant, ProteinModification):
                    continue
                elif variant.entity.name in UB_NAMES:
                    spia_matrices["inhibition_ubiquination"][u_name][v_name] = 1
                elif variant.entity.name in PH_NAMES:
                    spia_matrices["inhibition_phosphorylation"][u_name][v_name] = 1
        elif isinstance(v, (Gene, Rna)):  # Normal decrease, check which matrix
            spia_matrices["repression"][u_name][v_name] = 1
        else:
            spia_matrices["inhibition"][u_name][v_name] = 1

    elif relation == ASSOCIATION:
        spia_matrices["binding_association"][u_name][v_name] = 1


def spia_matrices_to_excel(spia_matrices: SPIADataFrames, path: str) -> None:
    """Export a SPIA data dictionary into an Excel sheet at the given path.

    .. note::

        # The R import should add the values:
        # ["nodes"] from the columns
        # ["title"] from the name of the file
        # ["NumberOfReactions"] set to "0"
    """
    writer = pd.ExcelWriter(path, engine="xlsxwriter")

    for relation, df in spia_matrices.items():
        df.to_excel(writer, sheet_name=relation, index=False)

    # Save excel
    writer.save()


def spia_matrices_to_tsvs(spia_matrices: SPIADataFrames, directory: str) -> None:
    """Export a SPIA data dictionary into a directory as several TSV documents."""
    os.makedirs(directory, exist_ok=True)
    for relation, df in spia_matrices.items():
        df.to_csv(
            os.path.join(directory, "{relation}.tsv".format(relation=relation)),
            index=True,
        )



# -*- coding: utf-8 -*-

"""Exceptions for input/output."""

from typing import Tuple

from ..exceptions import PyBELWarning

import_version_message_fmt = "Tried importing from PyBEL v{}. Need at least v{}"


class ImportVersionWarning(PyBELWarning, ValueError):
    """Raised when trying to import data from an old version of PyBEL."""

    def __init__(
        self,
        actual_version_tuple: Tuple[int, int, int],
        minimum_version_tuple: Tuple[int, int, int],
    ) -> None:
        """Build an import version warning."""
        super().__init__(actual_version_tuple, minimum_version_tuple)
        self.actual_tuple = actual_version_tuple
        self.minimum_tuple = minimum_version_tuple

    def __str__(self):
        actual_s = ".".join(map(str, self.actual_tuple))
        minimum_s = ".".join(map(str, self.minimum_tuple))
        return import_version_message_fmt.format(actual_s, minimum_s)



# -*- coding: utf-8 -*-

"""Predicates for edge data from BEL graphs."""

from functools import wraps
from typing import Any, Callable, Optional

from .typing import EdgePredicate
from .utils import part_has_modifier
from ..graph import BELGraph
from ...constants import (
    ACTIVITY,
    ANNOTATIONS,
    ASSOCIATION,
    CAUSAL_RELATIONS,
    CITATION,
    CITATION_AUTHORS,
    DEGRADATION,
    DIRECT_CAUSAL_RELATIONS,
    EVIDENCE,
    NAMESPACE,
    POLAR_RELATIONS,
    RELATION,
    SOURCE_MODIFIER,
    TARGET_MODIFIER,
    TRANSLOCATION,
)
from ...dsl import BaseEntity, BiologicalProcess, Pathology
from ...typing import EdgeData

__all__ = [
    "edge_predicate",
    "true_edge_predicate",
    "false_edge_predicate",
    "has_provenance",
    "has_pubmed",
    "has_pmc",
    "has_authors",
    "is_causal_relation",
    "not_causal_relation",
    "is_direct_causal_relation",
    "is_associative_relation",
    "has_polarity",
    "edge_has_activity",
    "edge_has_degradation",
    "edge_has_translocation",
    "edge_has_annotation",
    "has_pathology_causal",
]

DictEdgePredicate = Callable[[EdgeData], bool]


def edge_predicate(func: DictEdgePredicate) -> EdgePredicate:  # noqa: D202
    """Decorate an edge predicate function that only takes a dictionary as its singular argument.

    Apply this as a decorator to a function that takes a single argument, a PyBEL node data dictionary, to make
    sure that it can also accept a pair of arguments, a BELGraph and a PyBEL node tuple as well.
    """

    @wraps(func)
    def _wrapped(*args):
        x = args[0]

        if isinstance(x, BELGraph):
            u, v, k = args[1:4]
            return func(x[u][v][k])

        return func(*args)

    return _wrapped


def true_edge_predicate(graph: BELGraph, u: BaseEntity, v: BaseEntity, k: str) -> bool:
    """Return true for all edges."""
    return True


def false_edge_predicate(graph: BELGraph, u: BaseEntity, v: BaseEntity, k: str) -> bool:
    """Return false for all edges."""
    return False


@edge_predicate
def has_provenance(edge_data: EdgeData) -> bool:
    """Check if the edge has provenance information (i.e. citation and evidence)."""
    return CITATION in edge_data and EVIDENCE in edge_data


@edge_predicate
def has_pubmed(edge_data: EdgeData) -> bool:
    """Check if the edge has a PubMed citation."""
    return CITATION in edge_data and edge_data[CITATION][NAMESPACE].lower() in (
        "pubmed",
        "pmid",
    )


@edge_predicate
def has_pmc(edge_data: EdgeData) -> bool:
    """Check if the edge has a PMC citation."""
    return CITATION in edge_data and edge_data[CITATION][NAMESPACE].lower() in (
        "pmc",
        "pmcid",
    )


CITATION_PREDICATES = {
    "pubmed": has_pubmed,
    "pmc": has_pmc,
}


@edge_predicate
def has_authors(edge_data: EdgeData) -> bool:
    """Check if the edge contains author information for its citation."""
    return CITATION in edge_data and CITATION_AUTHORS in edge_data[CITATION] and edge_data[CITATION][CITATION_AUTHORS]


@edge_predicate
def is_causal_relation(edge_data: EdgeData) -> bool:
    """Check if the given relation is causal."""
    return edge_data[RELATION] in CAUSAL_RELATIONS


@edge_predicate
def not_causal_relation(edge_data: EdgeData) -> bool:
    """Check if the given relation is not causal."""
    return edge_data[RELATION] not in CAUSAL_RELATIONS


@edge_predicate
def is_direct_causal_relation(edge_data: EdgeData) -> bool:
    """Check if the edge is a direct causal relation."""
    return edge_data[RELATION] in DIRECT_CAUSAL_RELATIONS


@edge_predicate
def is_associative_relation(edge_data: EdgeData) -> bool:
    """Check if the edge has an association relation."""
    return edge_data[RELATION] == ASSOCIATION


@edge_predicate
def has_polarity(edge_data: EdgeData) -> bool:
    """Check if the edge has polarity."""
    return edge_data[RELATION] in POLAR_RELATIONS


def _has_modifier(edge_data: EdgeData, modifier: str) -> bool:
    """Check if the edge has the given modifier.

    :param edge_data: The edge data dictionary
    :param modifier: The modifier to check. One of :data:`pybel.constants.ACTIVITY`,
                        :data:`pybel.constants.DEGRADATION`, or :data:`pybel.constants.TRANSLOCATION`.
    :return: Does either the subject or object have the given modifier
    """
    return part_has_modifier(edge_data, SOURCE_MODIFIER, modifier) or part_has_modifier(
        edge_data, TARGET_MODIFIER, modifier
    )


@edge_predicate
def edge_has_activity(edge_data: EdgeData) -> bool:
    """Check if the edge contains an activity in either the subject or object."""
    return _has_modifier(edge_data, ACTIVITY)


@edge_predicate
def edge_has_translocation(edge_data: EdgeData) -> bool:
    """Check if the edge has a translocation in either the subject or object."""
    return _has_modifier(edge_data, TRANSLOCATION)


@edge_predicate
def edge_has_degradation(edge_data: EdgeData) -> bool:
    """Check if the edge contains a degradation in either the subject or object."""
    return _has_modifier(edge_data, DEGRADATION)


def edge_has_annotation(edge_data: EdgeData, key: str) -> Optional[Any]:
    """Check if an edge has the given annotation.

    :param edge_data: The data dictionary from a BELGraph's edge
    :param key: An annotation key
    :return: If the annotation key is present in the current data dictionary

    For example, it might be useful to print all edges that are annotated with 'Subgraph':

    >>> from pybel.examples import sialic_acid_graph
    >>> from pybel.examples.sialic_acid_example import sialic_acid_cd33_complex, cd33
    >>> edges = {
    ...     (u, v)
    ...     for u, v, data in sialic_acid_graph.edges(data=True)
    ...     if edge_has_annotation(data, 'Species')
    ... }
    >>> assert (sialic_acid_cd33_complex, cd33) in edges
    """
    annotations = edge_data.get(ANNOTATIONS)

    if annotations is None:
        return None

    return annotations.get(key)


def has_pathology_causal(graph: BELGraph, u: BaseEntity, v: BaseEntity, k: str) -> bool:
    """Check if the subject is a pathology and has a causal relationship with a non bioprocess/pathology.

    :return: If the subject of this edge is a pathology and it participates in a causal reaction.
    """
    return (
        isinstance(u, Pathology)
        and is_causal_relation(graph, u, v, k)
        and not isinstance(v, (Pathology, BiologicalProcess))
    )



# -*- coding: utf-8 -*-

"""Mutations for inferring new edges in the graph."""

from . import protein_rna_origins, transfer
from .protein_rna_origins import *
from .transfer import *

__all__ = [k for k in locals() if not k.startswith("_")]
